cdef extern from "esl_dirichlet.h" nogil:

    double esl_dirichlet_logpdf  (double *p, double *alpha, int K)
    double esl_dirichlet_logpdf_c(double *c, double *alpha, int K)

    # Sampling
    extern int esl_dirichlet_DSample       (ESL_RANDOMNESS *r, double *alpha, int K, double *p)
    extern int esl_dirichlet_FSample       (ESL_RANDOMNESS *r, float  *alpha, int K, float  *p)
    extern int esl_dirichlet_DSampleUniform(ESL_RANDOMNESS *r, int K, double *p)
    extern int esl_dirichlet_FSampleUniform(ESL_RANDOMNESS *r, int K, float  *p)
    extern int esl_dirichlet_SampleBeta    (ESL_RANDOMNESS *r, double theta1, double theta2, double *ret_answer)



from libc.stdint cimport uint8_t, uint32_t, int64_t

from libeasel.rand64 cimport ESL_RAND64
from libhmmer.p7_alidisplay cimport P7_ALIDISPLAY


cdef extern from "hmmer.h" nogil:

    ctypedef p7_dom_s P7_DOMAIN
    cdef struct p7_dom_s:
        int64_t ienv
        int64_t jenv
        int64_t iali
        int64_t jali
        int64_t iorf
        int64_t jorf
        float envsc
        float domcorrection
        float dombias
        float oasc
        float bitscore
        double lnP
        bint is_reported
        bint is_included
        float* scores_per_pos
        P7_ALIDISPLAY* ad

    P7_DOMAIN *p7_domain_Create_empty();
    void p7_domain_Destroy(P7_DOMAIN *obj);
    int p7_domain_Serialize(const P7_DOMAIN *obj, uint8_t **buf, uint32_t *n, uint32_t *nalloc);
    int p7_domain_Deserialize(const uint8_t *buf, uint32_t *n, P7_DOMAIN *ret_obj);
    int p7_domain_TestSample(ESL_RAND64 *rng, P7_DOMAIN **ret_obj);
    int p7_domain_Compare(P7_DOMAIN *first, P7_DOMAIN *second, double atol, double rtol);



cdef extern from "esl_hyperexp.h" nogil:

    ctypedef struct ESL_HYPEREXP:
        double* q
        double* lambda_
        double* wrk
        double  mu
        int     K
        char*   fixlambda
        int     fixmix

    ESL_HYPEREXP *esl_hyperexp_Create(int K);
    void          esl_hyperexp_Destroy(ESL_HYPEREXP *h);
    int           esl_hyperexp_Copy(ESL_HYPEREXP *src, ESL_HYPEREXP *dest);
    int           esl_hyperexp_FixedUniformMixture(ESL_HYPEREXP *h);
    int           esl_hyperexp_SortComponents(ESL_HYPEREXP *h);
    int           esl_hyperexp_Write(FILE *fp, ESL_HYPEREXP *hxp);
    int           esl_hyperexp_Dump(FILE *fp, ESL_HYPEREXP *hxp);

    int           esl_hyperexp_Read(ESL_FILEPARSER *ef, ESL_HYPEREXP **ret_hxp);
    int           esl_hyperexp_ReadFile(char *filename, ESL_HYPEREXP **ret_hxp);


    double  esl_hxp_pdf    (double x, ESL_HYPEREXP *h);
    double  esl_hxp_logpdf (double x, ESL_HYPEREXP *h);
    double  esl_hxp_cdf    (double x, ESL_HYPEREXP *h);
    double  esl_hxp_logcdf (double x, ESL_HYPEREXP *h);
    double  esl_hxp_surv   (double x, ESL_HYPEREXP *h);
    double  esl_hxp_logsurv(double x, ESL_HYPEREXP *h);
    double  esl_hxp_invcdf (double p, ESL_HYPEREXP *h);

    double  esl_hxp_generic_pdf   (double x, void *params);
    double  esl_hxp_generic_cdf   (double x, void *params);
    double  esl_hxp_generic_surv  (double x, void *params);
    double  esl_hxp_generic_invcdf(double x, void *params);

    int esl_hxp_Plot(FILE *fp, ESL_HYPEREXP *h,
    			double (*func)(double x, ESL_HYPEREXP *h),
    			double xmin, double xmax, double xstep);


    double esl_hxp_Sample(ESL_RANDOMNESS *r, ESL_HYPEREXP *h);

    int esl_hxp_FitGuess   (double *x, int n, ESL_HYPEREXP *h);
    int esl_hxp_FitComplete(double *x, int n, ESL_HYPEREXP *h);

    int esl_hxp_FitGuessBinned   (ESL_HISTOGRAM *g, ESL_HYPEREXP *h);
    int esl_hxp_FitCompleteBinned(ESL_HISTOGRAM *g, ESL_HYPEREXP *h);



# coding: utf-8
# cython: language_level=3, linetrace=True

# --- C imports --------------------------------------------------------------

from libc.stdint cimport uint32_t
from posix.types cimport off_t

from libeasel.sq cimport ESL_SQ
from libhmmer.p7_alidisplay cimport P7_ALIDISPLAY
from libhmmer.p7_bg cimport P7_BG
from libhmmer.p7_builder cimport P7_BUILDER
from libhmmer.p7_domain cimport P7_DOMAIN
from libhmmer.p7_hmm cimport P7_HMM
from libhmmer.p7_hmmfile cimport P7_HMMFILE
from libhmmer.p7_pipeline cimport P7_PIPELINE, p7_pipemodes_e
from libhmmer.p7_profile cimport P7_PROFILE
from libhmmer.p7_scoredata cimport P7_SCOREDATA
from libhmmer.p7_tophits cimport P7_TOPHITS, P7_HIT
from libhmmer.p7_trace cimport P7_TRACE

IF HMMER_IMPL == "VMX":
    from libhmmer.impl_vmx.p7_oprofile cimport P7_OPROFILE
ELIF HMMER_IMPL == "SSE":
    from libhmmer.impl_sse.p7_oprofile cimport P7_OPROFILE

from .easel cimport Alphabet, DigitalSequence, DigitalMSA, MSA, Randomness, VectorF


cdef extern from "hmmer.h" nogil:
    DEF p7_NOFFSETS = 3
    DEF p7_NEVPARAM = 6
    DEF p7_NCUTOFFS = 6


# --- Cython classes ---------------------------------------------------------


cdef class Alignment:
    cdef readonly Domain domain
    cdef P7_ALIDISPLAY* _ad


cdef class Background:
    cdef readonly bint     uniform
    cdef readonly Alphabet alphabet
    cdef readonly VectorF  residue_frequencies
    cdef          P7_BG*   _bg
    cdef          int      _L

    cpdef Background copy(self)


cdef class Builder:
    cdef readonly object prior_scheme
    cdef readonly object effective_number
    cdef readonly object architecture
    cdef readonly object weighting
    cdef public   str    score_matrix
    cdef public   double popen
    cdef public   double pextend
    cdef readonly Alphabet alphabet
    cdef readonly Randomness randomness

    cdef uint32_t    _seed    # the seed passed at builder initialization
    cdef P7_BUILDER* _bld

    cpdef tuple build(self, DigitalSequence sequence, Background background)
    cpdef tuple build_msa(self, DigitalMSA msa, Background background)
    cpdef Builder copy(self)


cdef class Cutoffs:
    cdef object              _owner
    cdef int*                _flags
    cdef bint                _is_profile
    cdef float[p7_NCUTOFFS]* _cutoffs

    cpdef VectorF as_vector(self)
    cpdef bint gathering_available(self)
    cpdef bint trusted_available(self)
    cpdef bint noise_available(self)


cdef class Domain:
    cdef readonly Alignment alignment
    cdef readonly Hit hit
    cdef P7_DOMAIN* _dom


cdef class Domains:
    cdef readonly Hit hit


cdef class EvalueParameters:
    cdef object              _owner
    cdef float[p7_NEVPARAM]* _evparams

    cpdef VectorF as_vector(self)


cdef class Hit:
    # a reference to the TopHits that owns the wrapped P7_HIT, kept so that
    # the internal data is never deallocated before the Python class.
    cdef readonly TopHits hits
    cdef P7_HIT* _hit

    cpdef bint is_included(self)
    cpdef bint is_reported(self)
    cpdef bint is_new(self)
    cpdef bint is_dropped(self)
    cpdef bint is_duplicate(self)


cdef class HMM:
    cdef P7_HMM* _hmm
    # a reference to the Alphabet Python object to avoid deallocation of the
    # inner ESL_ALPHABET; the Python object provides reference counting for free
    cdef readonly Alphabet          alphabet

    cpdef dict __getstate__(self)
    cpdef object __setstate__(self, dict state)

    cpdef HMM copy(self)
    cpdef VectorF match_occupancy(self)
    cpdef double mean_match_entropy(self) except *
    cpdef double mean_match_information(self, Background background) except *
    cpdef double mean_match_relative_entropy(self, Background background) except *
    cpdef void renormalize(self)
    cpdef void scale(self, double scale, bint exponential=?)
    cpdef void set_composition(self)
    cpdef void write(self, object fh, bint binary=*) except *
    cpdef void zero(self)


cdef class HMMFile:
    cdef P7_HMMFILE* _hfp
    cdef Alphabet _alphabet

    cpdef HMM read(self)
    cpdef void close(self)
    cpdef bint is_pressed(self)
    cpdef HMMPressedFile optimized_profiles(self)

    @staticmethod
    cdef P7_HMMFILE* _open_fileobj(object fh) except *


cdef class HMMPressedFile:
    cdef P7_HMMFILE* _hfp
    cdef Alphabet _alphabet
    cdef HMMFile _hmmfile

    cpdef OptimizedProfile read(self)


cdef class OptimizedProfile:
    cdef P7_OPROFILE* _om
    cdef readonly Alphabet alphabet

    cpdef OptimizedProfile copy(self)
    cpdef bint is_local(self)
    cpdef void write(self, object fh_filter, object fh_profile) except *

    # @staticmethod
    cdef int _convert(self, P7_PROFILE* gm) nogil except 1

    cpdef object ssv_filter(self, DigitalSequence seq)


cdef class Offsets:
    cdef object              _owner
    cdef off_t[p7_NOFFSETS]* _offs


cdef class Pipeline:
    cdef          object     _Z            # either `Z` as an int, or `None`
    cdef          object     _domZ         # either `domZ` as an int, or `None`
    cdef          uint32_t   _seed         # the seed passed at pipeline initialization
    cdef          void**     _refs         # the array to pass the references to the C code
    cdef          ssize_t    _nref         # the total size of `self._refs`
    cdef          dict       _cutoff_save  # a local save of the reporting parameters

    cdef readonly Alphabet         alphabet
    cdef readonly Background       background
    cdef readonly Profile          profile
    cdef readonly OptimizedProfile opt
    cdef readonly Randomness       randomness

    cdef OptimizedProfile _optimized
    cdef P7_PIPELINE* _pli

    cdef int _save_cutoff_parameters(self) except 1
    cdef int _restore_cutoff_parameters(self) except 1
    cdef P7_OPROFILE* _get_om_from_query(self, object query, int L = *) except NULL
    cpdef void    clear(self)
    cpdef TopHits search_hmm(self, object query, object seqs)
    cpdef TopHits search_msa(self, DigitalMSA query, object seqs, Builder builder = ?)
    cpdef TopHits search_seq(self, DigitalSequence query, object seqs, Builder builder = ?)
    @staticmethod
    cdef  int    _search_loop(
        P7_PIPELINE* pli,
        P7_OPROFILE* om,
        P7_BG*       bg,
        ESL_SQ**     seqs,
        P7_TOPHITS*  th,
    ) nogil except 1
    cpdef TopHits scan_seq(self, DigitalSequence query, object hmms)
    cdef int _scan_loop(
        self,
        P7_PIPELINE* pli,
        ESL_SQ*      sq,
        P7_BG*       bg,
        P7_HMM*      hm,
        P7_TOPHITS*  th,
        object       hmm_iter,
        Alphabet     hmm_alphabet
    ) except 1


cdef class LongTargetsPipeline(Pipeline):
    cdef DigitalSequence _tmpsq

    @staticmethod
    cdef int _search_loop_longtargets(
        P7_PIPELINE*  pli,
        P7_OPROFILE*  om,
        P7_BG*        bg,
        ESL_SQ**      sq,
        P7_TOPHITS*   th,
        P7_SCOREDATA* scoredata,
        ESL_SQ*       tmpsq,
    ) nogil except 1


cdef class Profile:
    cdef P7_PROFILE* _gm
    cdef readonly Alphabet          alphabet

    cdef int _clear(self) nogil except 1
    cdef int _configure(self, HMM hmm, Background background, int L, bint multihit=*, bint local=*) nogil except 1

    cpdef Profile copy(self)
    cpdef bint is_local(self)
    cpdef bint is_multihit(self)
    cpdef OptimizedProfile optimized(self)


cdef class ScoreData:
    cdef P7_SCOREDATA* _sd
    cdef readonly int  Kp

    cpdef ScoreData copy(self)


cdef class TopHits:
    cdef readonly float Z
    cdef readonly float domZ
    cdef readonly bint  long_targets

    cdef P7_TOPHITS* _th

    cdef int _threshold(self, Pipeline pipeline) nogil except 1
    cdef int _sort_by_key(self) nogil except 1
    cdef int _sort_by_seqidx(self) nogil except 1

    cpdef MSA to_msa(self, Alphabet alphabet, bint trim=*, bint digitize=?, bint all_consensus_cols=?)


cdef class Trace:
    cdef readonly Traces traces
    cdef P7_TRACE* _tr

    cpdef float expected_accuracy(self)


cdef class Traces:
    cdef P7_TRACE** _traces
    cdef size_t     _ntraces


cdef class TraceAligner:
    cdef ESL_SQ** _seqs
    cdef size_t   _nseq

    cpdef Traces compute_traces(self, HMM hmm, object sequences)
    cpdef MSA align_traces(
        self,
        HMM hmm,
        object sequences,
        Traces traces,
        bint trim=*,
        bint digitize=*,
        bint all_consensus_cols=*
    )



from libc.stdint cimport uint8_t, int64_t


cdef extern from "stdarg.h" nogil:
    ctypedef struct va_list:
          pass

cdef extern from "easel.h" nogil:

    cdef size_t eslERRBUFSIZE

    cdef double eslCONST_E
    cdef double eslCONST_PI
    cdef double eslCONST_EULER
    cdef double eslCONST_GOLD
    cdef double eslCONST_LOG2
    cdef double eslCONST_LOG2R
    cdef double eslINFINITY

    ctypedef uint8_t ESL_DSQ
    cdef     ESL_DSQ eslDSQ_SENTINEL
    cdef     ESL_DSQ eslDSQ_ILLEGAL
    cdef     ESL_DSQ eslDSQ_IGNORED
    cdef     ESL_DSQ eslDSQ_EOL
    cdef     ESL_DSQ eslDSQ_EOD

    ctypedef int64_t esl_pos_t

    cdef enum:
        eslOK = 0
        eslFAIL = 1
        eslEOL = 2
        eslEOF = 3
        eslEOD = 4
        eslEMEM = 5
        eslENOTFOUND = 6
        eslEFORMAT = 7
        eslEAMBIGUOUS = 8
        eslEDIVZERO = 9
        eslEINCOMPAT = 10
        eslEINVAL = 11
        eslESYS = 12
        eslECORRUPT = 13
        eslEINCONCEIVABLE = 14
        eslESYNTAX = 15
        eslERANGE = 16
        eslEDUP = 17
        eslENOHALT = 18
        eslENORESULT = 19
        eslENODATA = 20
        eslETYPE = 21
        eslEOVERWRITE = 22
        eslENOSPACE = 23
        eslEUNIMPLEMENTED = 24
        eslENOFORMAT = 25
        eslENOALPHABET = 26
        eslEWRITE = 27
        eslEINACCURATE = 28

    # 1. Error handling
    ctypedef void(*esl_exception_handler_f)(int errcode, int use_errno, char *sourcefile, int sourceline, char *format, va_list argp)
    # void esl_fail(char *errbuf, const char *format, ...)
    # void esl_exception(int errcode, int use_errno, char *sourcefile, int sourceline, char *format, ...);
    void esl_exception_SetHandler(esl_exception_handler_f)
    void esl_exception_ResetDefaultHandler()
    # void esl_nonfatal_handler(int errcode, int use_errno, char *sourcefile, int sourceline, char *format, va_list argp);
    # void esl_fatal(const char *format, ...) ESL_ATTRIBUTE_NORETURN;

    # 4. Improved replacements for some C library functions
    # extern int  esl_fgets(char **buf, int *n, FILE *fp);
    # extern int  esl_fprintf(FILE *fp, const char *format, ...);
    # extern int  esl_printf(const char *format, ...);
    # extern int  esl_strdup(const char *s, int64_t n, char **ret_dup);
    # extern int  esl_strcat(char **dest, int64_t ldest, const char *src, int64_t lsrc);
    # extern int  esl_strmapcat        (const ESL_DSQ *inmap, char **dest, int64_t *ldest, const char *src, esl_pos_t lsrc);
    # extern int  esl_strmapcat_noalloc(const ESL_DSQ *inmap,  char *dest, int64_t *ldest, const char *src, esl_pos_t lsrc);
    # extern int  esl_strtok    (char **s, char *delim, char **ret_tok);
    # extern int  esl_strtok_adv(char **s, char *delim, char **ret_tok, int *opt_toklen, char *opt_endchar);
    # extern int  esl_sprintf (char **ret_s, const char *format, ...);
    # extern int  esl_vsprintf(char **ret_s, const char *format, va_list *ap);
    # extern int  esl_strcmp(const char *s1, const char *s2);



cdef extern from "hmmer.h" nogil:
    int   p7_FLogsumInit()
    float p7_FLogsum(float a, float b)
    float p7_FLogsumError(float a, float b)
    int   p7_ILogsumInit()
    int   p7_ILogsum(int s1, int s2)



from libc.stdint cimport int64_t

from libeasel cimport ESL_DSQ, esl_pos_t


cdef extern from "esl_alphabet.h" nogil:

    cdef enum:
        eslUNKNOWN = 0
        eslRNA = 1
        eslDNA = 2
        eslAMINO = 3
        eslCOINS = 4
        eslDICE = 5
        eslNONSTANDARD = 6

    ctypedef struct ESL_ALPHABET:
        int type
        int K
        int Kp
        char* sym
        ESL_DSQ[128] inmap
        char** degen
        int* ndegen
        ESL_DSQ* complement

    # ESL alphabet object
    ESL_ALPHABET *esl_alphabet_Create(int type);
    ESL_ALPHABET *esl_alphabet_CreateCustom(const char *alphabet, int K, int Kp);
    int           esl_alphabet_SetEquiv(ESL_ALPHABET *a, char sym, char c);
    int           esl_alphabet_SetCaseInsensitive(ESL_ALPHABET *a);
    int           esl_alphabet_SetDegeneracy(ESL_ALPHABET *a, char c, char *ds);
    int           esl_alphabet_SetIgnored(ESL_ALPHABET *a, const char *ignoredchars);
    size_t        esl_alphabet_Sizeof(ESL_ALPHABET *a);
    void          esl_alphabet_Destroy(ESL_ALPHABET *a);

    # Digitized sequences
    int     esl_abc_CreateDsq(const ESL_ALPHABET *a, const char    *seq,        ESL_DSQ **ret_dsq);
    int     esl_abc_Digitize (const ESL_ALPHABET *a, const char    *seq,        ESL_DSQ *dsq);
    int     esl_abc_Textize  (const ESL_ALPHABET *a, const ESL_DSQ *dsq,  int64_t L, char   *seq);
    int     esl_abc_TextizeN (const ESL_ALPHABET *a, const ESL_DSQ *dptr, int64_t L, char   *buf);
    int     esl_abc_dsqcpy(const ESL_DSQ *dsq, int64_t L, ESL_DSQ *dcopy);
    int     esl_abc_dsqdup(const ESL_DSQ *dsq, int64_t L, ESL_DSQ **ret_dup);
    int     esl_abc_dsqcat        (const ESL_DSQ *inmap, ESL_DSQ **dsq, int64_t *L, const char *s, esl_pos_t n);
    int     esl_abc_dsqcat_noalloc(const ESL_DSQ *inmap, ESL_DSQ  *dsq, int64_t *L, const char *s, esl_pos_t n);
    int64_t esl_abc_dsqlen(const ESL_DSQ *dsq);
    int64_t esl_abc_dsqrlen(const ESL_ALPHABET *a, const ESL_DSQ *dsq);
    int     esl_abc_CDealign(const ESL_ALPHABET *abc, char    *s, const ESL_DSQ *ref_ax, int64_t *opt_rlen);
    int     esl_abc_XDealign(const ESL_ALPHABET *abc, ESL_DSQ *x, const ESL_DSQ *ref_ax, int64_t *opt_rlen);
    int     esl_abc_ConvertDegen2X(const ESL_ALPHABET *abc, ESL_DSQ *dsq);
    int     esl_abc_revcomp(const ESL_ALPHABET *abc, ESL_DSQ *dsq, int n);

    # Other routines in the API
    int    esl_abc_ValidateType(int type);
    int    esl_abc_GuessAlphabet(const int64_t *ct, int *ret_type);
    double esl_abc_Match       (const ESL_ALPHABET *a, ESL_DSQ x, ESL_DSQ y, double *p);
    int    esl_abc_IAvgScore   (const ESL_ALPHABET *a, ESL_DSQ x, const int    *sc);
    float  esl_abc_FAvgScore   (const ESL_ALPHABET *a, ESL_DSQ x, const float  *sc);
    double esl_abc_DAvgScore   (const ESL_ALPHABET *a, ESL_DSQ x, const double *sc);
    int    esl_abc_IExpectScore(const ESL_ALPHABET *a, ESL_DSQ x, const int    *sc, const float  *p);
    float  esl_abc_FExpectScore(const ESL_ALPHABET *a, ESL_DSQ x, const float  *sc, const float  *p);
    double esl_abc_DExpectScore(const ESL_ALPHABET *a, ESL_DSQ x, const double *sc, const double *p);

    int    esl_abc_IAvgScVec   (const ESL_ALPHABET *a, int    *sc);
    int    esl_abc_FAvgScVec   (const ESL_ALPHABET *a, float  *sc);
    int    esl_abc_DAvgScVec   (const ESL_ALPHABET *a, double *sc);
    int    esl_abc_IExpectScVec(const ESL_ALPHABET *a, int    *sc, const float  *p);
    int    esl_abc_FExpectScVec(const ESL_ALPHABET *a, float  *sc, const float  *p);
    int    esl_abc_DExpectScVec(const ESL_ALPHABET *a, double *sc, const double *p);
    int    esl_abc_FCount      (const ESL_ALPHABET *a, float  *ct, ESL_DSQ x, float  wt);
    int    esl_abc_DCount      (const ESL_ALPHABET *a, double *ct, ESL_DSQ x, double wt);
    int    esl_abc_EncodeType   (char *typestring);
    int    esl_abc_EncodeTypeMem(char *type, int n);
    char  *esl_abc_DecodeType   (int type);
    int    esl_abc_ValidateSeq(const ESL_ALPHABET *a, const char *seq, int64_t L, char *errbuf);

    ESL_DSQ esl_abc_DigitizeSymbol (ESL_ALPHABET*, char)

    bint    esl_abc_XIsValid       (ESL_ALPHABET*, int)
    bint    esl_abc_XIsResidue     (ESL_ALPHABET*, int)
    bint    esl_abc_XIsCanonical   (ESL_ALPHABET*, int)
    bint    esl_abc_XIsGap         (ESL_ALPHABET*, int)
    bint    esl_abc_XIsDegenerate  (ESL_ALPHABET*, int)
    bint    esl_abc_XIsUnknown     (ESL_ALPHABET*, int)
    bint    esl_abc_XIsNonresidue  (ESL_ALPHABET*, int)
    bint    esl_abc_XIsMissing     (ESL_ALPHABET*, int)
    int     esl_abc_XGetGap        (ESL_ALPHABET*)
    int     esl_abc_XGetUnknown    (ESL_ALPHABET*)
    int     esl_abc_XGetNonresidue (ESL_ALPHABET*)
    int     esl_abc_XGetMissing    (ESL_ALPHABET*)

    bint    esl_abc_CIsValid       (ESL_ALPHABET*, char)
    bint    esl_abc_CIsResidue     (ESL_ALPHABET*, char)
    bint    esl_abc_CIsCanonical   (ESL_ALPHABET*, char)
    bint    esl_abc_CIsGap         (ESL_ALPHABET*, char)
    bint    esl_abc_CIsDegenerate  (ESL_ALPHABET*, char)
    bint    esl_abc_CIsUnknown     (ESL_ALPHABET*, char)
    bint    esl_abc_CIsNonresidue  (ESL_ALPHABET*, char)
    bint    esl_abc_CIsMissing     (ESL_ALPHABET*, char)
    char    esl_abc_CGetGap        (ESL_ALPHABET*)
    char    esl_abc_CGetUnknown    (ESL_ALPHABET*)
    char    esl_abc_CGetNonresidue (ESL_ALPHABET*)
    char    esl_abc_CGetMissing    (ESL_ALPHABET*)



from libc.stdio cimport FILE

from libhmmer.impl_sse.p7_oprofile cimport P7_OPROFILE


cdef extern from "emmintrin.h":

    ctypedef struct __m128i:
        pass

    ctypedef struct __m128:
        pass


cdef extern from "impl_sse/impl_sse.h" nogil:

    ctypedef struct P7_OM_BLOCK:
        int count
        int listSize
        P7_OPROFILE** list

    ctypedef p7_omx_s P7_OMX
    ctypedef struct p7_omx_s:
        int       M
        int       L

        __m128  **dpf
        __m128i **dpw
        __m128i **dpb
        void     *dp_mem
        int       allocR
        int       validR
        int       allocQ4
        int       allocQ8
        int       allocQ16
        size_t    ncells

        float    *xmx
        void     *x_mem
        int       allocXR
        float     totscale
        bint       has_own_scales

        bint     debugging;
        FILE   *dfp;


    P7_OMX* p7_omx_Create(int allocM, int allocL, int allocXL)
    int     p7_omx_GrowTo(P7_OMX *ox, int allocM, int allocL, int allocXL)
    int     p7_omx_Reuse(P7_OMX *ox)
    void    p7_omx_Destroy(P7_OMX *ox)



from libc.stdint cimport uint64_t, int64_t
from libc.stdio cimport FILE

from libeasel.sq cimport ESL_SQ
from libeasel.getopts cimport ESL_GETOPTS
from libeasel.random cimport ESL_RANDOMNESS
from libhmmer.fm cimport FM_DATA, FM_CFG
from libhmmer.p7_bg cimport P7_BG
from libhmmer.p7_domaindef cimport P7_DOMAINDEF
from libhmmer.p7_tophits cimport P7_TOPHITS
from libhmmer.p7_hmmfile cimport P7_HMMFILE
from libhmmer.p7_scoredata cimport P7_SCOREDATA

IF HMMER_IMPL == "VMX":
    from libhmmer.impl_vmx.p7_omx cimport P7_OMX
    from libhmmer.impl_vmx.p7_oprofile cimport P7_OPROFILE
ELIF HMMER_IMPL == "SSE":
    from libhmmer.impl_sse.p7_omx cimport P7_OMX
    from libhmmer.impl_sse.p7_oprofile cimport P7_OPROFILE


cdef extern from "easel.h" nogil:

    DEF eslERRBUFSIZE = 128


cdef extern from "hmmer.h" nogil:

    cdef enum p7_pipemodes_e:
        p7_SEARCH_SEQS = 0
        p7_SCAN_MODELS = 1

    cdef enum p7_zsetby_e:
        p7_ZSETBY_NTARGETS = 0
        p7_ZSETBY_OPTION = 1
        p7_ZSETBY_FILEINFO = 2

    cdef enum p7_complementarity_e:
        p7_NOCOMPLEMENT = 0
        p7_COMPLEMENT = 1

    cdef enum p7_strands_e:
        p7_STRAND_TOPONLY = 0
        p7_STRAND_BOTTOMONLY = 1
        p7_STRAND_BOTH = 2

    # ctypedef p7_pipeline_s P7_PIPELINE
    ctypedef struct P7_PIPELINE:
        P7_OMX* oxf
        P7_OMX* oxb
        P7_OMX* fwd
        P7_OMX* bck

        ESL_RANDOMNESS* r
        bint do_reseeding
        bint do_alignment_score_calc
        P7_DOMAINDEF* ddef

        int by_E
        double E
        double T
        bint dom_by_E
        double domE
        double domT
        int use_bit_cutoffs

        bint inc_by_E
        double incE
        double incT
        bint incdom_by_E
        double incdomE
        double incdomT

        double Z
        double domZ
        p7_zsetby_e Z_setby
        p7_zsetby_e domZ_setby

        bint do_max
        double F1
        double F2
        double F3
        int B1
        int B2
        int B3
        bint do_biasfilter
        bint do_null2

        uint64_t nmodels
        uint64_t nseqs
        uint64_t nres
        uint64_t nnodes
        uint64_t n_past_msv
        uint64_t n_past_bias
        uint64_t n_past_vit
        uint64_t n_past_fwd
        uint64_t n_output
        uint64_t pos_past_msv
        uint64_t pos_past_bias
        uint64_t pos_past_vit
        uint64_t pos_past_fwd
        uint64_t pos_output

        p7_pipemodes_e mode
        bint long_targets
        p7_strands_e strands
        int W
        int block_length

        bint show_accessions
        bint show_alignments

        P7_HMMFILE   *hfp
        char          errbuf[eslERRBUFSIZE];


    P7_PIPELINE *p7_pipeline_Create(const ESL_GETOPTS *go, int M_hint, int L_hint, int long_targets, p7_pipemodes_e mode)
    int          p7_pipeline_Reuse  (P7_PIPELINE *pli)
    void         p7_pipeline_Destroy(P7_PIPELINE *pli)
    int          p7_pipeline_Merge  (P7_PIPELINE *p1, P7_PIPELINE *p2)

    # int p7_pli_ExtendAndMergeWindows (P7_OPROFILE *om, const P7_SCOREDATA *msvdata, P7_HMM_WINDOWLIST *windowlist, float pct_overlap)
    int p7_pli_TargetReportable   (P7_PIPELINE *pli, float score,     double lnP)
    int p7_pli_DomainReportable  (P7_PIPELINE *pli, float dom_score, double lnP)

    int p7_pli_TargetIncludable  (P7_PIPELINE *pli, float score,     double lnP)
    int p7_pli_DomainIncludable  (P7_PIPELINE *pli, float dom_score, double lnP)
    int p7_pli_NewModel          (P7_PIPELINE *pli, const P7_OPROFILE *om, P7_BG *bg)
    int p7_pli_NewModelThresholds(P7_PIPELINE *pli, const P7_OPROFILE *om)
    int p7_pli_NewSeq            (P7_PIPELINE *pli, const ESL_SQ *sq)
    int p7_Pipeline              (P7_PIPELINE *pli, P7_OPROFILE *om, P7_BG *bg, const ESL_SQ *sq, const ESL_SQ *ntsq, P7_TOPHITS *th)
    int p7_Pipeline_LongTarget   (
        P7_PIPELINE* pli,
        P7_OPROFILE* om,
        P7_SCOREDATA* data,
        P7_BG* bg,
        P7_TOPHITS* hitlist,
        int64_t seqidx,
        const ESL_SQ* sq,
        p7_complementarity_e complementarity,
        const FM_DATA* fmf,
        const FM_DATA* fmb,
        FM_CFG* fm_cfg
    )

    int p7_pli_Statistics(FILE* ofp, P7_PIPELINE* pli, void* w)



cdef extern from "esl_cpu.h" nogil:

    bint   esl_cpu_has_sse(void);
    bint   esl_cpu_has_sse4(void);
    bint   esl_cpu_has_avx(void);
    bint   esl_cpu_has_avx512(void);
    char *esl_cpu_Get(void);



from libc.stdio cimport FILE
from libc.stdint cimport int16_t, int32_t, int64_t, uint16_t, uint32_t, uint64_t
from posix.types cimport off_t


cdef extern from "easel.h":

    DEF eslERRBUFSIZE = 128


cdef extern from "esl_ssi.h" nogil:

    cdef int eslSSI_MAXFILES
    cdef int eslSSI_MAXKEYS
    cdef int eslSSI_MAXRAM


    ctypedef struct ESL_SSI:
        FILE* fp
        uint32_t flags
        uint32_t offsz
        uint16_t nfiles
        uint64_t nprimary
        uint64_t nsecondary
        uint32_t flen
        uint32_t plen
        uint32_t slen
        uint32_t frecsize
        uint32_t precsize
        uint32_t srecsize
        off_t foffset
        off_t poffset
        off_t soffset
        char** filename
        uint32_t* fileformat
        uint32_t* fileflags
        uint32_t* bpl
        uint32_t* rpl


    ctypedef struct ESL_PKEY:
        char* key
        uint16_t fnum
        off_t r_off
        off_t d_off
        int64_t len


    ctypedef struct ESL_SKEY:
        char* key
        char* pkey


    ctypedef struct ESL_NEWSSI:
        char* ssifile
        FILE* ssifp
        int external
        int max_ram

        char** filename
        uint32_t* fileformat
        uint32_t* bpl
        uint32_t* rpl
        uint32_t flen
        uint16_t nfiles

        ESL_PKEY* pkeys
        uint32_t plen
        uint64_t nprimary
        char* ptmpfile
        FILE* ptmp

        ESL_SKEY* skeys
        uint32_t slen
        uint64_t nsecondary
        char* stmpfile
        FILE* stmp

        char errbuf[eslERRBUFSIZE]


    # Using (reading) SSI indices
    int  esl_ssi_Open(const char *filename, ESL_SSI **ret_ssi)
    void esl_ssi_Close(ESL_SSI *ssi)
    int  esl_ssi_FindName(ESL_SSI *ssi, const char *key,
    			     uint16_t *ret_fh, off_t *ret_roff, off_t *opt_doff, int64_t *opt_L)
    int  esl_ssi_FindNumber(ESL_SSI *ssi, int64_t nkey,
    			       uint16_t *opt_fh, off_t *opt_roff, off_t *opt_doff, int64_t *opt_L, char **opt_pkey)
    int  esl_ssi_FindSubseq(ESL_SSI *ssi, const char *key, int64_t requested_start,
    			       uint16_t *ret_fh, off_t *ret_roff, off_t *ret_doff, int64_t *ret_L, int64_t *ret_actual_start)
    int  esl_ssi_FileInfo(ESL_SSI *ssi, uint16_t fh, char **ret_filename, int *ret_format)



    # Creating (writing) SSI indices.
    int  esl_newssi_Open(const char *ssifile, int allow_overwrite, ESL_NEWSSI **ret_newssi)
    int  esl_newssi_AddFile  (ESL_NEWSSI *ns, const char *filename, int fmt, uint16_t *ret_fh)
    int  esl_newssi_SetSubseq(ESL_NEWSSI *ns, uint16_t fh, uint32_t bpl, uint32_t rpl)
    int  esl_newssi_AddKey   (ESL_NEWSSI *ns, const char *key, uint16_t fh, off_t r_off, off_t d_off, int64_t L)
    int  esl_newssi_AddAlias (ESL_NEWSSI *ns, const char *alias, const char *key)
    int  esl_newssi_Write    (ESL_NEWSSI *ns)
    void esl_newssi_Close    (ESL_NEWSSI *ns)


    # Portable binary i/o.
    void     esl_byteswap(char *swap, int nbytes)
    uint16_t esl_ntoh16(uint16_t netshort)
    uint32_t esl_ntoh32(uint32_t netlong)
    uint64_t esl_ntoh64(uint64_t net_int64)
    uint16_t esl_hton16(uint16_t hostshort)
    uint32_t esl_hton32(uint32_t hostlong)
    uint64_t esl_hton64(uint64_t host_int64)
    int      esl_fread_u16(FILE *fp, uint16_t *ret_result)
    int      esl_fread_u32(FILE *fp, uint32_t *ret_result)
    int      esl_fread_u64(FILE *fp, uint64_t *ret_result)
    int      esl_fread_i16(FILE *fp, int16_t  *ret_result)
    int      esl_fread_i32(FILE *fp, int32_t  *ret_result)
    int      esl_fread_i64(FILE *fp, int64_t  *ret_result)
    int      esl_fwrite_u16(FILE *fp, uint16_t n)
    int      esl_fwrite_u32(FILE *fp, uint32_t n)
    int      esl_fwrite_u64(FILE *fp, uint64_t n)
    int      esl_fwrite_i16(FILE *fp, int16_t  n)
    int      esl_fwrite_i32(FILE *fp, int32_t  n)
    int      esl_fwrite_i64(FILE *fp, int64_t  n)
    int	esl_fread_offset(FILE *fp, int mode, off_t *ret_offset)
    int      esl_fwrite_offset(FILE *fp, off_t offset)



from libc.stdint cimport uint64_t

from libeasel cimport ESL_DSQ
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.random cimport ESL_RANDOMNESS


cdef extern from "esl_hmm.h" nogil:

    ctypedef struct ESL_HMM:
        int M
        int K
        float* pi
        float** t
        float** e
        float** eo
        const ESL_ALPHABET* abc


    ctypedef struct ESL_HMX:
        float** dp
        float* sc
        int M
        int L
        float* dp_mem
        int allocR
        int validR
        int allocM
        uint64_t ncells


    ESL_HMM *esl_hmm_Create(const ESL_ALPHABET *abc, int M);
    ESL_HMM *esl_hmm_Clone(const ESL_HMM *hmm);
    int      esl_hmm_Configure(ESL_HMM *hmm, float *fq);
    int      esl_hmm_SetDegeneracies(ESL_HMM *hmm);
    void     esl_hmm_Destroy(ESL_HMM *hmm);

    ESL_HMX *esl_hmx_Create(int allocL, int allocM);
    int      esl_hmx_GrowTo (ESL_HMX *mx, int L, int M);
    void     esl_hmx_Destroy(ESL_HMX *mx);

    int      esl_hmm_Emit(ESL_RANDOMNESS *r, const ESL_HMM *hmm, ESL_DSQ **opt_dsq, int **opt_path, int *opt_L);
    int      esl_hmm_Forward(const ESL_DSQ *dsq, int L, const ESL_HMM *hmm, ESL_HMX *fwd, float *opt_sc);
    int      esl_hmm_Backward(const ESL_DSQ *dsq, int L, const ESL_HMM *hmm, ESL_HMX *bck, float *opt_sc);



# coding: utf-8
import typing

statuscode: typing.Dict[int, str]

class UnexpectedError(RuntimeError):
    code: int
    function: str
    def __init__(self, code: int, function: str) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...

class AllocationError(MemoryError):
    ctype: str
    itemsize: int
    count: int
    def __init__(self, ctype: str, itemsize: int, count: int = 1) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...

class EaselException(RuntimeError):
    code: int
    message: str
    def __init__(self, code: int, message: str) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...



from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.mixdchlet cimport ESL_MIXDCHLET
from libhmmer.p7_hmm cimport P7_HMM

cdef extern from "hmmer.h" nogil:

    ctypedef p7_prior_s P7_PRIOR
    cdef struct p7_prior_s:
        ESL_MIXDCHLET* tm
        ESL_MIXDCHLET* ti
        ESL_MIXDCHLET* td
        ESL_MIXDCHLET* em
        ESL_MIXDCHLET* ei

    P7_PRIOR  *p7_prior_CreateAmino()
    P7_PRIOR  *p7_prior_CreateNucleic()
    P7_PRIOR  *p7_prior_CreateLaplace(const ESL_ALPHABET *abc)
    void       p7_prior_Destroy(P7_PRIOR *pri)

    int        p7_ParameterEstimation(P7_HMM *hmm, const P7_PRIOR* pri)



from libc.stdio cimport FILE

from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.fileparser cimport ESL_FILEPARSER
from libeasel.dmatrix cimport ESL_DMATRIX


cdef extern from "esl_scorematrix.h" nogil:

    ctypedef struct ESL_SCOREMATRIX:
        int** s
        int   K
        int   Kp

        char* isval
        const ESL_ALPHABET* abc_r

        int nc
        char* outorder

        char* name
        char* path

    ESL_SCOREMATRIX *esl_scorematrix_Create(const ESL_ALPHABET *abc)
    int              esl_scorematrix_Copy(const ESL_SCOREMATRIX *src, ESL_SCOREMATRIX *dest)
    ESL_SCOREMATRIX *esl_scorematrix_Clone(const ESL_SCOREMATRIX *S)
    int              esl_scorematrix_Compare(const ESL_SCOREMATRIX *S1, const ESL_SCOREMATRIX *S2)
    int              esl_scorematrix_CompareCanon(const ESL_SCOREMATRIX *S1, const ESL_SCOREMATRIX *S2)
    int              esl_scorematrix_Max(const ESL_SCOREMATRIX *S)
    int              esl_scorematrix_Min(const ESL_SCOREMATRIX *S)
    int              esl_scorematrix_IsSymmetric(const ESL_SCOREMATRIX *S)
    int              esl_scorematrix_ExpectedScore(ESL_SCOREMATRIX *S, double *fi, double *fj, double *ret_E)
    int              esl_scorematrix_RelEntropy(const ESL_SCOREMATRIX *S, const double *fi, const double *fj, double lambda_, double *ret_D)
    int              esl_scorematrix_JointToConditionalOnQuery(const ESL_ALPHABET *abc, ESL_DMATRIX *P)
    void             esl_scorematrix_Destroy(ESL_SCOREMATRIX *S)

    int              esl_scorematrix_Set(const char *name, ESL_SCOREMATRIX *S)
    int              esl_scorematrix_SetIdentity(ESL_SCOREMATRIX *S)

    int              esl_scorematrix_SetFromProbs(ESL_SCOREMATRIX *S, double lambda_, const ESL_DMATRIX *P, const double *fi, const double *fj)
    int              esl_scorematrix_SetWAG(ESL_SCOREMATRIX *S, double lambda_, double t)

    int  esl_scorematrix_Read(ESL_FILEPARSER *efp, const ESL_ALPHABET *abc, ESL_SCOREMATRIX **ret_S)
    int  esl_scorematrix_Write(FILE *fp, const ESL_SCOREMATRIX *S)

    int esl_scorematrix_ProbifyGivenBG(const ESL_SCOREMATRIX *S, const double *fi, const double *fj, double *opt_lambda, ESL_DMATRIX **opt_P)

    int esl_scorematrix_Probify(const ESL_SCOREMATRIX *S, ESL_DMATRIX **opt_P, double **opt_fi, double **opt_fj, double *opt_lambda)



from libhmmer.p7_profile cimport P7_PROFILE

IF HMMER_IMPL == "VMX":
    from libhmmer.impl_vmx.p7_oprofile cimport P7_OPROFILE
ELIF HMMER_IMPL == "SSE":
    from libhmmer.impl_sse.p7_oprofile cimport P7_OPROFILE

cdef extern from "hmmer.h" nogil:

    cdef enum p7_scoredatatype_e:
        p7_sd_std = 0
        p7_sd_fm = 1

    cdef struct p7_scoredata_s:
        p7_scoredatatype_e type
        int M
        float* prefix_lengths
        float* suffix_lengths
        float* fwd_scores
        float** fwd_transitions
        float** opt_ext_fwd
        float** opt_ext_rev
    ctypedef p7_scoredata_s P7_SCOREDATA

    cdef void p7_hmm_ScoreDataDestroy(P7_SCOREDATA* data)
    cdef P7_SCOREDATA* p7_hmm_ScoreDataCreate(P7_OPROFILE* om, P7_PROFILE* gm)
    cdef P7_SCOREDATA* p7_hmm_ScoreDataClone(P7_SCOREDATA* src, int Kp)
    cdef int p7_hmm_ScoreDataComputeRest(P7_OPROFILE* om, P7_SCOREDATA* data)



from libc.stdio cimport FILE

from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.msa cimport ESL_MSA
from libeasel.sq cimport ESL_SQ
from libhmmer.p7_profile cimport P7_PROFILE
from libhmmer.p7_hmm cimport P7_HMM
from libhmmer.p7_gmx cimport P7_GMX
from libhmmer.p7_trace cimport P7_TRACE


cdef extern from "hmmer.h" nogil:

  int p7_tracealign_Seqs(ESL_SQ **sq, P7_TRACE **tr, int nseq, int M, int optflags, P7_HMM *hmm, ESL_MSA **ret_msa)
  int p7_tracealign_MSA(const ESL_MSA *premsa, P7_TRACE **tr, int M, int optflags, ESL_MSA **ret_postmsa)
  int p7_tracealign_computeTraces(P7_HMM *hmm, ESL_SQ  **sq, int offset, int N, P7_TRACE  **tr)
  int p7_tracealign_getMSAandStats(P7_HMM *hmm, ESL_SQ  **sq, int N, ESL_MSA **ret_msa, float **ret_pp, float **ret_relent, float **ret_scores )



from libc.stdint cimport uint32_t
from libc.stdio cimport FILE
from posix.types cimport off_t

from libeasel cimport ESL_DSQ
from libeasel.sq cimport ESL_SQ
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.random cimport ESL_RANDOMNESS


cdef extern from "hmmer.h" nogil:

    DEF p7_NEVPARAM = 6
    cdef size_t p7_NEVPARAM

    DEF p7_NCUTOFFS = 6
    cdef size_t p7_NCUTOFFS

    DEF p7_NOFFSETS = 3
    cdef size_t p7_NOFFSETS

    DEF p7_MAXABET = 20
    cdef size_t p7_MAXABET

    cdef size_t p7H_NTRANSITIONS = 7
    cdef enum p7h_transitions_e:
        p7H_MM = 0
        p7H_MI = 1
        p7H_MD = 2
        p7H_IM = 3
        p7H_II = 4
        p7H_DM = 5
        p7H_DD = 6

    cdef enum:
        p7H_HASBITS = (1<<0)
        p7H_DESC    = (1<<1)
        p7H_RF      = (1<<2)
        p7H_CS      = (1<<3)
        p7H_XRAY    = (1<<4)
        p7H_HASPROB = (1<<5)
        p7H_HASDNA  = (1<<6)
        p7H_STATS   = (1<<7)
        p7H_MAP     = (1<<8)
        p7H_ACC     = (1<<9)
        p7H_GA      = (1<<10)
        p7H_TC      = (1<<11)
        p7H_NC      = (1<<12)
        p7H_CA      = (1<<13)
        p7H_COMPO   = (1<<14)
        p7H_CHKSUM  = (1<<15)
        p7H_CONS    = (1<<16)
        p7H_MMASK   = (1<<17)

    ctypedef p7_hmm_s P7_HMM
    cdef struct p7_hmm_s:

        int M
        float** t
        float** mat
        float** ins

        char* name
        char* acc
        char* desc
        char* rf
        char* mm
        char* consensus
        char* cs
        char* ca

        char* comlog
        int nseq
        float eff_nseq
        int max_length
        char* ctime
        int* map
        uint32_t checksum
        float[p7_NEVPARAM] evparam
        float[p7_NCUTOFFS] cutoff
        float[p7_MAXABET] compo

        off_t offset
        ESL_ALPHABET* abc
        int flags


    # The P7_HMM object: allocation, initialization, destruction.
    P7_HMM *p7_hmm_Create(int M, const ESL_ALPHABET *abc)
    P7_HMM *p7_hmm_CreateShell()
    int     p7_hmm_CreateBody(P7_HMM *hmm, int M, const ESL_ALPHABET *abc)
    void    p7_hmm_Destroy(P7_HMM *hmm)
    int     p7_hmm_CopyParameters(const P7_HMM *src, P7_HMM *dest)
    P7_HMM *p7_hmm_Clone(const P7_HMM *hmm)
    int     p7_hmm_Zero(P7_HMM *hmm)
    char    p7_hmm_EncodeStatetype(char *typestring)
    char   *p7_hmm_DecodeStatetype(char st)
    # Convenience routines for setting fields in an HMM.
    int     p7_hmm_SetName       (P7_HMM *hmm, char *name);
    int     p7_hmm_SetAccession  (P7_HMM *hmm, char *acc);
    int     p7_hmm_SetDescription(P7_HMM *hmm, char *desc);
    int     p7_hmm_AppendComlog  (P7_HMM *hmm, int argc, char **argv);
    int     p7_hmm_SetCtime      (P7_HMM *hmm);
    int     p7_hmm_SetComposition(P7_HMM *hmm);
    int     p7_hmm_SetConsensus  (P7_HMM *hmm, ESL_SQ *sq);
    # Renormalization and rescaling counts in core HMMs.
    int     p7_hmm_Scale      (P7_HMM *hmm, double scale);
    int     p7_hmm_ScaleExponential(P7_HMM *hmm, double exp);
    int     p7_hmm_Renormalize(P7_HMM *hmm);
    # Debugging and development code.
    int     p7_hmm_Dump(FILE *fp, P7_HMM *hmm);
    int     p7_hmm_Sample          (ESL_RANDOMNESS *r, int M, const ESL_ALPHABET *abc, P7_HMM **ret_hmm);
    int     p7_hmm_SampleUngapped  (ESL_RANDOMNESS *r, int M, const ESL_ALPHABET *abc, P7_HMM **ret_hmm);
    int     p7_hmm_SampleEnumerable(ESL_RANDOMNESS *r, int M, const ESL_ALPHABET *abc, P7_HMM **ret_hmm);
    int     p7_hmm_SampleUniform   (ESL_RANDOMNESS *r, int M, const ESL_ALPHABET *abc,
    				     float tmi, float tii, float tmd, float tdd,  P7_HMM **ret_hmm);
    int     p7_hmm_Compare(P7_HMM *h1, P7_HMM *h2, float tol);
    int     p7_hmm_Validate(P7_HMM *hmm, char *errbuf, float tol);
    # Other routines in the API
    int     p7_hmm_CalculateOccupancy(const P7_HMM *hmm, float *mocc, float *iocc);










  #
  # # struct p7_hmmfile_s:
  # #     FILE* f
  # #     char* filename
  # #     ESL_SSI* ssi
  # #
  # #     bint do_gzip
  # #     bint do_stdin
  # #     bint newly_opened
  # #     bint is_pressed
  # #
  # #     int format
  # #     int (*parser)(struct p7_hmmfile_s*, ESL_ALPHABET**, P7_HMM**)
  # #
  # # ctypedef p7_hmmfile_s P7_HMMFILE
  #
  #
  #
  # int p7_hmmfile_OpenE(const char *filename, char *env, P7_HMMFILE **ret_hfp, char *errbuf)



from libc.stdio cimport FILE


cdef extern from "esl_getopts.h" nogil:

    ctypedef struct ESL_OPTIONS:
        char* name
        int type
        char* defval
        char* envvar
        char* range
        char* toggle_opts
        char* required_opts
        char* incompat_opts
        char* help
        int docgrouptag


    ctypedef struct ESL_GETOPTS:
        ESL_OPTIONS* opt
        int nopts
        int argc
        char** argv
        int optind
        int nfiles
        char** val
        int* setby
        int* valloc
        char* optstring
        char* spoof
        char** spoof_argv


    ESL_GETOPTS *esl_getopts_Create(const ESL_OPTIONS *opt);
    ESL_GETOPTS *esl_getopts_CreateDefaultApp(const ESL_OPTIONS *options, int nargs, int argc, char **argv, char *banner, char *usage);
    int          esl_getopts_Reuse  (ESL_GETOPTS *g);
    void         esl_getopts_Destroy(ESL_GETOPTS *g);
    void         esl_getopts_Dump(FILE *ofp, ESL_GETOPTS *g);

    int esl_opt_ProcessConfigfile (ESL_GETOPTS *g, char *filename, FILE *fp);
    int esl_opt_ProcessEnvironment(ESL_GETOPTS *g);
    int esl_opt_ProcessCmdline    (ESL_GETOPTS *g, int argc, char **argv);
    int esl_opt_ProcessSpoof      (ESL_GETOPTS *g, const char *cmdline);
    int esl_opt_VerifyConfig      (ESL_GETOPTS *g);
    int esl_opt_ArgNumber   (const ESL_GETOPTS *g);
    int esl_opt_SpoofCmdline(const ESL_GETOPTS *g, char **ret_cmdline);

    int esl_opt_GetSetter(const ESL_GETOPTS *g, char *optname);

    int    esl_opt_IsDefault (const ESL_GETOPTS *g, char *optname);
    int    esl_opt_IsOn      (const ESL_GETOPTS *g, char *optname);
    int    esl_opt_IsUsed    (const ESL_GETOPTS *g, char *optname);

    int    esl_opt_GetBoolean(const ESL_GETOPTS *g, char *optname);
    int    esl_opt_GetInteger(const ESL_GETOPTS *g, char *optname);
    double esl_opt_GetReal   (const ESL_GETOPTS *g, char *optname);
    char   esl_opt_GetChar   (const ESL_GETOPTS *g, char *optname);
    char  *esl_opt_GetString (const ESL_GETOPTS *g, char *optname);
    char  *esl_opt_GetArg    (const ESL_GETOPTS *g, int which);

    int esl_opt_DisplayHelp(FILE *ofp, const ESL_GETOPTS *go, int docgroup, int indent, int textwidth);



from libc.stdio cimport FILE

from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.msa cimport ESL_MSA
from libeasel.sq cimport ESL_DSQ
from libhmmer.p7_profile cimport P7_PROFILE
from libhmmer.p7_hmm cimport P7_HMM
from libhmmer.p7_gmx cimport P7_GMX


cdef extern from "hmmer.h" nogil:

    DEF p7T_NSTATETYPES = 12
    cdef enum p7t_statetype_e:
        p7T_BOGUS =  0
        p7T_M     =  1
        p7T_D     =  2
        p7T_I     =  3
        p7T_S     =  4
        p7T_N     =  5
        p7T_B     =  6
        p7T_E     =  7
        p7T_C     =  8
        p7T_T     =  9
        p7T_J     = 10
        p7T_X     = 11


    ctypedef p7_trace_s P7_TRACE
    cdef struct p7_trace_s:
        int    N
        int    nalloc
        char  *st
        int   *k
        int   *i
        float *pp
        int    M
        int    L
        int   ndom
        int  *tfrom
        int  *tto
        int  *sqfrom
        int  *sqto
        int  *hmmfrom
        int  *hmmto
        int   ndomalloc


    P7_TRACE *p7_trace_Create()
    P7_TRACE *p7_trace_CreateWithPP()
    int  p7_trace_Reuse(P7_TRACE *tr)
    int  p7_trace_Grow(P7_TRACE *tr)
    int  p7_trace_GrowIndex(P7_TRACE *tr)
    int  p7_trace_GrowTo(P7_TRACE *tr, int N)
    int  p7_trace_GrowIndexTo(P7_TRACE *tr, int ndom)
    void p7_trace_Destroy(P7_TRACE *tr)
    void p7_trace_DestroyArray(P7_TRACE **tr, int N)

    int  p7_trace_GetDomainCount   (const P7_TRACE *tr, int *ret_ndom)
    int  p7_trace_GetStateUseCounts(const P7_TRACE *tr, int *counts)
    int  p7_trace_GetDomainCoords  (const P7_TRACE *tr, int which, int *ret_i1, int *ret_i2, int *ret_k1, int *ret_k2)

    int   p7_trace_Validate(const P7_TRACE *tr, const ESL_ALPHABET *abc, const ESL_DSQ *dsq, char *errbuf)
    int   p7_trace_Dump(FILE *fp, const P7_TRACE *tr, const P7_PROFILE *gm, const ESL_DSQ *dsq)
    int   p7_trace_Compare(P7_TRACE *tr1, P7_TRACE *tr2, float pptol)
    int   p7_trace_Score(P7_TRACE *tr, ESL_DSQ *dsq, P7_PROFILE *gm, float *ret_sc)
    int   p7_trace_SetPP(P7_TRACE *tr, const P7_GMX *pp)
    float p7_trace_GetExpectedAccuracy(const P7_TRACE *tr)

    int  p7_trace_Append(P7_TRACE *tr, char st, int k, int i)
    int  p7_trace_AppendWithPP(P7_TRACE *tr, char st, int k, int i, float pp)
    int  p7_trace_Reverse(P7_TRACE *tr)
    int  p7_trace_Index(P7_TRACE *tr)

    int  p7_trace_FauxFromMSA(ESL_MSA *msa, int *matassign, int optflags, P7_TRACE **tr)
    int  p7_trace_Doctor(P7_TRACE *tr, int *opt_ndi, int *opt_nid)

    int  p7_trace_Count(P7_HMM *hmm, ESL_DSQ *dsq, float wt, P7_TRACE *tr)



cdef extern from "esl_mem.h" nogil:
    int       esl_mem_strtoi32(char *p, esl_pos_t n, int base, int *opt_nc, int32_t *opt_val);
    int       esl_mem_strtoi64(char *p, esl_pos_t n, int base, int *opt_nc, int64_t *opt_val);
    int       esl_mem_strtoi  (char *p, esl_pos_t n, int base, int *opt_nc, int     *opt_val);
    int       esl_mem_strtof  (char *p, esl_pos_t n,           int *opt_nc, float   *opt_val);
    int       esl_memnewline(const char *p, esl_pos_t n, esl_pos_t *ret_nline, int *ret_nterm);
    int       esl_memtok(char **p, esl_pos_t *n, const char *delim, char **ret_tok, esl_pos_t *ret_toklen);
    esl_pos_t esl_memspn (char *p, esl_pos_t n, const char *allow);
    esl_pos_t esl_memcspn(char *p, esl_pos_t n, const char *disallow);
    int       esl_memstrcmp     (const char *p, esl_pos_t n, const char *s);
    int       esl_memstrcmp_case(const char *p, esl_pos_t n, const char *s);
    int       esl_memstrpfx     (const char *p, esl_pos_t n, const char *s);
    int       esl_memstrpfx_case(const char *p, esl_pos_t n, const char *s);
    int       esl_memstrcontains(const char *p, esl_pos_t n, const char *s);
    int       esl_memstrdup(const char *p, esl_pos_t n, char **ret_s);
    int       esl_memstrcpy(const char *p, esl_pos_t n, char *dest);
    int       esl_memtof(const char *p, esl_pos_t n, float  *ret_val);
    int       esl_memtod(const char *p, esl_pos_t n, double *ret_val);
    int       esl_mem_IsReal(const char *p, esl_pos_t n);



#!/usr/bin/env python
# coding: utf-8

import collections
import configparser
import glob
import os
import platform
import re
import sys
import subprocess
from unittest import mock

import setuptools
from distutils.command.clean import clean as _clean
from distutils.errors import CompileError
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.build_clib import build_clib as _build_clib
from setuptools.command.sdist import sdist as _sdist
from setuptools.extension import Extension, Library

try:
    from Cython.Build import cythonize
except ImportError as err:
    cythonize = err

# --- Utils ------------------------------------------------------------------

def _split_multiline(value):
    value = value.strip()
    sep = max('\n,;', key=value.count)
    return list(filter(None, map(lambda x: x.strip(), value.split(sep))))

def _eprint(*args, **kwargs):
    print(*args, **kwargs, file=sys.stderr)

# --- `setup.py` commands ----------------------------------------------------

class sdist(_sdist):
    """A `sdist` that generates a `pyproject.toml` on the fly.
    """

    def run(self):
        # build `pyproject.toml` from `setup.cfg`
        c = configparser.ConfigParser()
        c.add_section("build-system")
        c.set("build-system", "requires", str(self.distribution.setup_requires))
        c.set("build-system", 'build-backend', '"setuptools.build_meta"')
        with open("pyproject.toml", "w") as pyproject:
            c.write(pyproject)
        # run the rest of the packaging
        _sdist.run(self)


class build_ext(_build_ext):
    """A `build_ext` that disables optimizations if compiled in debug mode.
    """

    def finalize_options(self):
        _build_ext.finalize_options(self)
        self._clib_cmd = self.get_finalized_command("build_clib")
        self._clib_cmd.force = self.force
        self._clib_cmd.debug = self.debug

    def run(self):
        # check `cythonize` is available
        if isinstance(cythonize, ImportError):
            raise RuntimeError("Cython is required to run `build_ext` command") from cythonize

        # check the CPU architecture could be detected
        global machine
        if machine is None:
            raise RuntimeError("Could not detect CPU architecture with `platform.machine`")

        # check a platform-specific implementation of HMMER was selected
        # depending on the detected machine
        global hmmer_impl
        if hmmer_impl is None:
            raise RuntimeError('Could not select implementation for CPU architecture: "{}"'.format(machine))
        else:
            _eprint('Building HMMER with', hmmer_impl, 'for CPU architecture:', repr(machine))

        # use debug directives with Cython if building in debug mode
        cython_args = {
            "include_path": ["include"],
            "compiler_directives": {},
            "compile_time_env": {
                "SYS_IMPLEMENTATION_NAME": sys.implementation.name,
                "SYS_VERSION_INFO_MAJOR": sys.version_info.major,
                "SYS_VERSION_INFO_MINOR": sys.version_info.minor,
                "SYS_VERSION_INFO_MICRO": sys.version_info.micro,
            }
        }
        if hmmer_impl is not None:
            cython_args["compile_time_env"]["HMMER_IMPL"] = hmmer_impl
        if self.force:
            cython_args["force"] = True
        if self.debug:
            cython_args["gdb_debug"] = True
            cython_args["annotate"] = True
            cython_args["compiler_directives"]["warn.undeclared"] = True
            cython_args["compiler_directives"]["warn.unreachable"] = True
            cython_args["compiler_directives"]["warn.maybe_uninitialized"] = True
            cython_args["compiler_directives"]["warn.unused"] = True
            cython_args["compiler_directives"]["warn.unused_arg"] = True
            cython_args["compiler_directives"]["warn.unused_result"] = True
            cython_args["compiler_directives"]["warn.multiple_declarators"] = True
            cython_args["compiler_directives"]["profile"] = True
        else:
            cython_args["compiler_directives"]["boundscheck"] = False
            cython_args["compiler_directives"]["wraparound"] = False
            cython_args["compiler_directives"]["cdivision"] = True

        # cythonize and patch the extensions
        self.extensions = cythonize(self.extensions, **cython_args)
        for ext in self.extensions:
            ext._needs_stub = False

        # # update the compiler include and link dirs to use the
        # # temporary build folder so that the platform-specific headers
        # # and static libs can be found

        # check the libraries have been built already
        if not self.distribution.have_run["build_clib"]:
            self._clib_cmd.run()

        # build the extensions as normal
        _build_ext.run(self)

    def build_extension(self, ext):
        # update compile flags if compiling in debug mode
        if self.debug:
            if self.compiler.compiler_type in {"unix", "cygwin", "mingw32"}:
                ext.extra_compile_args.append("-Og")
                ext.extra_compile_args.append("--coverage")
                ext.extra_link_args.append("--coverage")
            elif self.compiler.compiler_type == "msvc":
                ext.extra_compile_args.append("/Od")
            if sys.implementation.name == "cpython":
                ext.define_macros.append(("CYTHON_TRACE_NOGIL", 1))
        else:
            ext.define_macros.append(("CYTHON_WITHOUT_ASSERTIONS", 1))

        # update link and include directories
        ext.include_dirs.append(self._clib_cmd.build_clib)
        ext.library_dirs.append(self._clib_cmd.build_clib)
        for name in ext.libraries:
            lib = self._clib_cmd.get_library(name)
            ext.include_dirs.extend(lib.include_dirs)
            ext.extra_objects.append(self.compiler.library_filename(
                lib.name, output_dir=self._clib_cmd.build_clib
            ))

        # build the rest of the extension as normal
        _build_ext.build_extension(self, ext)


class configure(_build_clib):
    """A `./configure`-style command to generate a platform-specific C header.

    Derives from the `build_clib` command from setuptools to be able to use
    the same configuration values (`self.build_temp`, `self.build_clib`, etc).
    """

    # --- Compatibility with base `build_clib` command ---

    def check_library_list(self, libraries):
        pass

    def get_library_names(self):
        return [ lib.name for lib in self.libraries ]

    def get_source_files(self):
        return [ source for lib in self.libraries for source in lib.sources ]

    # --- Autotools-like helpers ---

    def _silent_spawn(self, cmd):
        try:
            subprocess.run(cmd, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as err:
            raise CompileError(err.stderr)

    def _has_header(self, headername):
        _eprint('checking whether <{}> can be included'.format(headername), end="... ")

        slug = re.sub("[./-]", "_", headername)
        testfile = os.path.join(self.build_temp, "have_{}.c".format(slug))
        objects = []

        with open(testfile, "w") as f:
            f.write('#include "{}"\n'.format(headername))
        try:
            with mock.patch.object(self.compiler, "spawn", new=self._silent_spawn):
                objects = self.compiler.compile([testfile], debug=self.debug)
        except CompileError as err:
            _eprint("no")
            return False
        else:
            _eprint("yes")
            return True
        finally:
            os.remove(testfile)
            for obj in filter(os.path.isfile, objects):
                os.remove(obj)

    def _has_function(self, funcname):
        _eprint('checking whether function', repr(funcname), 'is available', end="... ")

        testfile = os.path.join(self.build_temp, "have_{}.c".format(funcname))
        binfile = os.path.join(self.build_temp, "have_{}.bin".format(funcname))
        objects = []

        with open(testfile, "w") as f:
            f.write('int main() {{return {}(); return 0;}}\n'.format(funcname))
        try:
            with mock.patch.object(self.compiler, "spawn", new=self._silent_spawn):
                objects = self.compiler.compile([testfile], debug=self.debug)
                self.compiler.link_executable(objects, binfile)
        except CompileError:
            _eprint("no")
            return False
        else:
            _eprint("yes")
            return True
        finally:
            os.remove(testfile)
            for obj in filter(os.path.isfile, objects):
                os.remove(obj)
            if os.path.isfile(binfile):
                os.remove(binfile)

    def _check_sse2(self):
        _eprint('checking whether compiler can build SSE2 code', end="... ")

        testfile = os.path.join(self.build_temp, "have_sse2.c")
        binfile = os.path.join(self.build_temp, "have_sse2.bin")
        objects = []

        with open(testfile, "w") as f:
            f.write("""
                #include <emmintrin.h>
                int main() {
                    __m128i a = _mm_set1_epi16(1);
                    short   x = _mm_extract_epi16(a, 1);
                    return (x == 1) ? 0 : 1;
                }
            """)
        try:
            with mock.patch.object(self.compiler, "spawn", new=self._silent_spawn):
                objects = self.compiler.compile([testfile], debug=self.debug, extra_preargs=["-msse2"])
                self.compiler.link_executable(objects, binfile)
                subprocess.run([binfile], check=True)
        except CompileError:
            _eprint("no")
            return False
        except subprocess.CalledProcessError:
            _eprint("yes, but cannot run code")
            return True  # assume we are cross-compiling, and still build
        else:
            _eprint("yes, with -msse2")
            return True
        finally:
            os.remove(testfile)
            for obj in filter(os.path.isfile, objects):
                os.remove(obj)
            if os.path.isfile(binfile):
                os.remove(binfile)

    def _check_vmx(self):
        _eprint('checking whether compiler can build VMX code', end="... ")

        testfile = os.path.join(self.build_temp, "have_sse.c")
        binfile = os.path.join(self.build_temp, "have_sse.bin")
        objects = []

        with open(testfile, "w") as f:
            f.write("""
                #include <altivec.h>
                int main() {
                    vector float a = vec_splats(1.0);
                    float f;
                    vec_ste(a, 0, &f);
                    return (f == 1) ? 0 : 1;
                }
            """)
        try:
            with mock.patch.object(self.compiler, "spawn", new=self._silent_spawn):
                objects = self.compiler.compile([testfile], debug=self.debug)
                self.compiler.link_executable(objects, binfile)
                subprocess.run([binfile], check=True)
        except CompileError:
            _eprint('no')
            return False
        except subprocess.CalledProcessError:
            _eprint('yes, but cannot run code')
            return True  # assume we are cross-compiling, and still build
        else:
            _eprint('yes')
            return True
        finally:
            os.remove(testfile)
            for obj in filter(os.path.isfile, objects):
                os.remove(obj)
            if os.path.isfile(binfile):
                os.remove(binfile)

    # --- Required interface for `setuptools.Command` ---

    def build_libraries(self, libraries):
        # read `setup.cfg`, which should be next to this file, so we can
        # use the `__file__` magic variable to locate it
        _cfg = configparser.ConfigParser()
        _cfg.read([__file__.replace(".py", ".cfg")])

        # ensure the output directory exists, otherwise create it
        self.mkpath(self.build_clib)

        # run the `configure_library` method sequentially on each library,
        # unless the header already exists and the configuration has not
        # been edited
        for library in self.distribution.libraries:
            section = "configure.{}".format(library.name)
            config = dict(_cfg.items(section))

            headers = _split_multiline(config.get("headers", ""))
            functions = _split_multiline(config.get("functions", ""))
            constants = dict(
                map(str.strip, line.split("="))
                for line in _split_multiline(config.get("constants", ""))
            )

            self.make_file(
                [__file__.replace(".py", ".cfg")],
                os.path.join(self.build_clib, config["filename"]),
                self.configure_library,
                (
                    library,
                    config["filename"],
                    config.get("copy"),
                    constants,
                    headers,
                    functions,
                ),
                exec_msg='generating "{}" for {} library'.format(
                    config["filename"], library.name
                )
            )

    def configure_library(self, library, filename, copy=None, constants=None, headers=None, functions=None):
        # if we only need to copy the header (e.g. for divsufsort) then just
        # do that and then exit
        if copy is not None:
            self.copy_file(copy, os.path.join(self.build_clib, filename))
            return

        # create a mapping to store the defines, and make sure it is ordered
        # (this is to keep compatibility with Python 3.5, a dict would do fine)
        defines = collections.OrderedDict()

        # fill the defines with the constants
        constants = constants or {}
        for k, v in constants.items():
            defines[k] = v

        # check endianness
        if sys.byteorder == "big":
            defines["WORDS_BIGENDIAN"] = 1

        # check platform flags
        global hmmer_impl
        platform_supported = False
        if hmmer_impl is not None:
            if hmmer_impl == "SSE":
                supported_feature = self._check_sse2()
            elif hmmer_impl == "VMX":
                supported_feature = self._check_vmx()
            else:
                supported_feature = False
            if not supported_feature:
                raise RuntimeError("failed to compile platform-specific code, aborting.")

        # fill the defines if headers are found
        headers = headers or []
        for header in headers:
            if self._has_header(header):
                slug = re.sub("[./-]", "_", header).upper()
                defines["HAVE_{}".format(slug)] = 1

        # fill the defines if functions are found
        functions = functions or []
        for func in functions:
            if self._has_function(func):
                defines["HAVE_{}".format(func.upper())] = 1

        # write the header file
        slug = re.sub("[./-]", "_", filename).upper()
        with open(os.path.join(self.build_clib, filename), "w") as f:
            f.write("#ifndef {}_INCLUDED\n".format(slug))
            f.write("#define {}_INCLUDED\n".format(slug))
            for k, v in defines.items():
                f.write("#define {} {}\n".format(k, '' if v is None else v))
            f.write("#endif\n".format(slug))


class build_clib(_build_clib):
    """A custom `build_clib` that compiles out of source.
    """

    # --- Distutils command interface ---

    def finalize_options(self):
        _build_clib.finalize_options(self)
        self._configure_cmd = self.get_finalized_command("configure")
        self._configure_cmd.force = self.force

    # --- Compatibility with base `build_clib` command ---

    def check_library_list(self, libraries):
        pass

    def get_library_names(self):
        return [ lib.name for lib in self.libraries ]

    def get_source_files(self):
        return [ source for lib in self.libraries for source in lib.sources ]

    # --- Helpers ---

    def get_library(self, name):
        return next(lib for lib in self.libraries if lib.name == name)

    # --- Build code ---

    def run(self):
        # make sure the C headers were generated already
        if not self.distribution.have_run["configure"]:
            self._configure_cmd.run()

        # patch the `esl_sqio_ascii.c` so we can use functions it otherwise
        # declares as `static`
        libeasel = self.get_library("easel")
        old = next(src for src in libeasel.sources if src.endswith("esl_sqio_ascii.c"))
        new = os.path.join(self.build_temp, "esl_sqio_ascii.c")
        self.make_file([old], new, self.destatic, (old, new))
        libeasel.sources.remove(old)
        libeasel.sources.append(new)

        # patch the `p7_hmmfile.c` so that we can use functions it otherwise
        # declares as `static`
        libhmmer = self.get_library("hmmer")
        old = next(src for src in libhmmer.sources if src.endswith("p7_hmmfile.c"))
        new = os.path.join(self.build_temp, "p7_hmmfile.c")
        self.make_file([old], new, self.destatic, (old, new))
        libhmmer.sources.remove(old)
        libhmmer.sources.append(new)

        # build the libraries normally
        _build_clib.run(self)

    def build_libraries(self, libraries):
        self.mkpath(self.build_clib)
        for library in libraries:
            self.make_file(
                library.sources,
                self.compiler.library_filename(library.name, output_dir=self.build_clib),
                self.build_library,
                (library,)
            )

    def build_library(self, library):
        # update compile flags if compiling in debug or release mode
        if self.debug:
            if self.compiler.compiler_type in {"unix", "cygwin", "mingw32"}:
                library.extra_compile_args.append("-Og")
                library.extra_compile_args.append("--coverage")
                library.extra_link_args.append("--coverage")
            elif self.compiler.compiler_type == "msvc":
                library.extra_compile_args.append("/Od")

        # build objects and create a static library
        objects = self.compiler.compile(
            library.sources,
            output_dir=self.build_temp,
            include_dirs=library.include_dirs + [self.build_clib],
            macros=library.define_macros,
            debug=self.debug,
            depends=library.depends,
            extra_preargs=library.extra_compile_args,
        )
        self.compiler.create_static_lib(
            objects,
            library.name,
            output_dir=self.build_clib,
            debug=self.debug,
        )

    def destatic(self, old, new):
        with open(old, "r") as f:
            lines = f.readlines()
        with open(new, "w") as f:
            f.writelines(l.replace("static ", "") for l in lines)


class clean(_clean):

    def remove_file(self, filename):
        if os.path.exists(filename):
            _eprint("removing", repr(filename))
            os.remove(filename)
        else:
            _eprint(repr(filename), "does not exist -- can't clean it")

    def run(self):
        _clean.run(self)

        _build_cmd = self.get_finalized_command("build_ext")
        _build_cmd.inplace = True

        for ext in self.distribution.ext_modules:
            filename = _build_cmd.get_ext_filename(ext.name)
            if self.all:
                self.remove_file(filename)
            basename = _build_cmd.get_ext_fullname(ext.name).replace(".", os.path.sep)
            for ext in ["c", "html"]:
                filename = os.path.extsep.join([basename, ext])
                self.remove_file(filename)


# --- C static libraries -----------------------------------------------------

# fmt: off
hmmer_sources = [
    os.path.join("vendor", "hmmer", "src", basename)
    for basename in [
        "build.c", "cachedb.c", "cachedb_shard.c", "emit.c", "errors.c",
    	"evalues.c", "eweight.c", "generic_decoding.c", "generic_fwdback.c",
    	"generic_fwdback_chk.c", "generic_fwdback_banded.c", "generic_null2.c",
    	"generic_msv.c", "generic_optacc.c", "generic_stotrace.c",
        "generic_viterbi.c", "generic_vtrace.c", "h2_io.c", "heatmap.c",
    	"hmmlogo.c", "hmmdmstr.c", "hmmdmstr_shard.c", "hmmd_search_status.c",
        "hmmdwrkr.c", "hmmdwrkr_shard.c", "hmmdutils.c", "hmmer.c", "logsum.c",
        "modelconfig.c", "modelstats.c", "mpisupport.c", "seqmodel.c",
        "tracealign.c", "p7_alidisplay.c", "p7_bg.c", "p7_builder.c",
        "p7_domain.c", "p7_domaindef.c", "p7_gbands.c", "p7_gmx.c",
        "p7_gmxb.c", "p7_gmxchk.c", "p7_hit.c", "p7_hmm.c", "p7_hmmcache.c",
    	"p7_hmmd_search_stats.c", "p7_hmmfile.c", "p7_hmmwindow.c",
    	"p7_pipeline.c", "p7_prior.c", "p7_profile.c", "p7_spensemble.c",
    	"p7_tophits.c", "p7_trace.c", "p7_scoredata.c", "hmmpgmd2msa.c",
    	"fm_alphabet.c", "fm_general.c", "fm_sse.c", "fm_ssv.c",
    ]
]

# HMMER3 is only supported on x86 CPUs with SSE, and big endian PowerPC
# (see https://github.com/EddyRivasLab/hmmer/issues/142)
machine = platform.machine().lower()
if machine.startswith('ppc') and not machine.endswith('le'):
    hmmer_sources.extend(glob.glob(os.path.join("vendor", "hmmer", "src", "impl_vmx", "*.c")))
    hmmer_sources.remove(os.path.join("vendor", "hmmer", "src", "impl_vmx", "vitscore.c"))
    hmmer_impl = "VMX"
    platform_define_macros = [("eslENABLE_VMX", 1)]
    platform_compile_args = ["-maltivec"]
elif machine.startswith(("x86", "amd", "i386", "i686")):
    hmmer_sources.extend(glob.glob(os.path.join("vendor", "hmmer", "src", "impl_sse", "*.c")))
    hmmer_sources.remove(os.path.join("vendor", "hmmer", "src", "impl_sse", "vitscore.c"))
    hmmer_impl = "SSE"
    platform_define_macros = [("eslENABLE_SSE", 1)]
    platform_compile_args = ["-msse3"]
else:
    _eprint('pyHMMER is not supported on CPU architecture:', repr(machine))
    platform_define_macros = []
    platform_compile_args = []
    hmmer_impl = None

libraries = [
    Library(
        "divsufsort",
        sources=[os.path.join("vendor", "hmmer", "libdivsufsort", "divsufsort.c")],
    ),
    Library(
        "easel",
        sources=glob.glob(os.path.join("vendor", "easel", "*.c")),
        include_dirs=[os.path.join("vendor", "easel")],
        define_macros=platform_define_macros,
        extra_compile_args=platform_compile_args,
    ),
    Library(
        "hmmer",
        sources=hmmer_sources,
        extra_compile_args=platform_compile_args,
        define_macros=platform_define_macros,
        include_dirs=[
            os.path.join("vendor", "easel"),
            os.path.join("vendor", "hmmer", "src"),
            os.path.join("vendor", "hmmer", "libdivsufsort"),
        ],
    ),
]


# --- Cython extensions ------------------------------------------------------

extensions = [
    Extension(
        "pyhmmer.errors",
        [os.path.join("pyhmmer", "errors.pyx")],
        libraries=["easel"],
    ),
    Extension(
        "pyhmmer.easel",
        [os.path.join("pyhmmer", "easel.pyx")],
        libraries=["easel"],
        define_macros=platform_define_macros,
        extra_compile_args=platform_compile_args,
    ),
    Extension(
        "pyhmmer.plan7",
        [os.path.join("pyhmmer", "plan7.pyx")],
        libraries=["hmmer", "easel", "divsufsort"],
        define_macros=platform_define_macros,
        extra_compile_args=platform_compile_args,
    ),
]


# --- Setup ------------------------------------------------------------------

setuptools.setup(
    ext_modules=extensions,
    libraries=libraries,
    cmdclass=dict(
        build_ext=build_ext,
        build_clib=build_clib,
        clean=clean,
        configure=configure,
        sdist=sdist,
    ),
)



# coding: utf-8
import abc
import array
import collections.abc
import os
import sys
import types
import typing

# --- Alphabet ---------------------------------------------------------------

class Alphabet(object):
    @classmethod
    def dna(self) -> Alphabet: ...
    @classmethod
    def rna(self) -> Alphabet: ...
    @classmethod
    def amino(self) -> Alphabet: ...
    def __repr__(self) -> str: ...
    def __eq__(self, other: object) -> bool: ...
    def __reduce__(self) -> typing.Tuple[typing.Callable[[], Alphabet], typing.Tuple[()]]: ...
    def __sizeof__(self) -> int: ...
    @property
    def K(self) -> int: ...
    @property
    def Kp(self) -> int: ...
    @property
    def symbols(self) -> str: ...
    def is_dna(self) -> bool: ...
    def is_rna(self) -> bool: ...
    def is_amino(self) -> bool: ...
    def is_nucleotide(self) -> bool: ...

# --- Bitfield ---------------------------------------------------------------

class Bitfield(typing.Sequence[bool]):
    def __init__(self, length: int) -> None: ...
    def __len__(self) -> int: ...
    @typing.overload
    def __getitem__(self, index: int) -> bool: ...
    @typing.overload
    def __getitem__(self, index: slice) -> typing.Sequence[bool]: ...
    def __setitem__(self, index: int, value: bool) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> typing.Dict[str, object]: ...
    def __setstate__(self, state: typing.Dict[str, object]) -> None: ...
    def __sizeof__(self) -> int: ...
    def count(self, value: bool = True) -> int: ...
    def toggle(self, index: int) -> None: ...

# --- KeyHash ----------------------------------------------------------------

class KeyHash(typing.Mapping[bytes, int]):
    def __init__(self) -> None: ...
    def __copy__(self) -> KeyHash: ...
    def __len__(self) -> int: ...
    def __contains__(self, value: object) -> bool: ...
    def __getitem__(self, item: bytes) -> int: ...
    def __iter__(self) -> typing.Iterator[bytes]: ...
    def __getstate__(self) -> typing.Dict[str, object]: ...
    def __setstate__(self, state: typing.Dict[str, object]) -> None: ...
    def __sizeof__(self) -> int: ...
    def clear(self) -> None: ...
    def copy(self) -> KeyHash: ...

# --- Matrix & Vector --------------------------------------------------------

_T = typing.TypeVar("_T")
_V = typing.TypeVar("_V")
_M = typing.TypeVar("_M")

class Vector(typing.Sequence[_T], typing.Generic[_T], abc.ABC):
    def __reduce__(self: _V) -> typing.Tuple[typing.Type[_V], typing.Tuple[array.array[typing.Any]]]: ...
    @abc.abstractmethod
    @classmethod
    def zeros(cls: typing.Type[_V], n: int) -> _V: ...
    @abc.abstractmethod
    def __init__(self, iterable: typing.Collection[_T]): ...
    def __copy__(self: _V) -> _V: ...
    @abc.abstractmethod
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    @abc.abstractmethod
    def __getitem__(self, index: int) -> _T: ...
    @typing.overload
    @abc.abstractmethod
    def __getitem__(self, index: slice) -> Vector[_T]: ...
    @typing.overload
    @abc.abstractmethod
    def __setitem__(self, index: int, value: _T) -> None: ...
    @typing.overload
    @abc.abstractmethod
    def __setitem__(self, index: slice, value: _T) -> None: ...
    @abc.abstractmethod
    def __iadd__(self: _V, other: typing.Union[_T, _V]) -> _V: ...
    @abc.abstractmethod
    def __imul__(self: _V, other: typing.Union[_T, _V]) -> _V: ...
    @abc.abstractmethod
    def __matmul__(self: _V, other: _V) -> _T: ...
    def __sizeof__(self) -> int: ...
    def __repr__(self) -> str: ...
    def __len__(self) -> int: ...
    @property
    def shape(self) -> typing.Tuple[int]: ...
    @property
    def strides(self) -> typing.Tuple[int]: ...
    @property
    @abc.abstractmethod
    def itemsize(self) -> int: ...
    @property
    @abc.abstractmethod
    def format(self) -> str: ...
    @abc.abstractmethod
    def argmax(self) -> int: ...
    @abc.abstractmethod
    def argmin(self) -> int: ...
    @abc.abstractmethod
    def copy(self: _V) -> _V: ...
    @abc.abstractmethod
    def max(self) -> _T: ...
    @abc.abstractmethod
    def min(self) -> _T: ...
    @abc.abstractmethod
    def reverse(self) -> None: ...
    @abc.abstractmethod
    def sum(self) -> _T: ...

class VectorF(Vector[float]):
    @classmethod
    def zeros(cls, n: int) -> VectorF: ...
    def __init__(self, iterable: typing.Optional[typing.Collection[float]] = None): ...
    def __copy__(self) -> VectorF: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> float: ...
    @typing.overload
    def __getitem__(self, index: slice) -> VectorF: ...
    @typing.overload
    def __setitem__(self, index: int, value: float) -> None: ...
    @typing.overload
    def __setitem__(self, index: slice, value: float) -> None: ...
    def __neg__(self) -> VectorU8: ...
    def __add__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __iadd__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __sub__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __isub__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __mul__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __imul__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __truediv__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __itruediv__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __matmul__(self, other: VectorF) -> float: ...
    def argmax(self) -> int: ...
    def argmin(self) -> int: ...
    def copy(self) -> VectorF: ...
    def entropy(self) -> float: ...
    def max(self) -> float: ...
    def min(self) -> float: ...
    def normalize(self) -> None: ...
    def relative_entropy(self, other: VectorF) -> float: ...
    def reverse(self) -> None: ...
    def sum(self) -> float: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

class VectorU8(Vector[int]):
    @classmethod
    def zeros(cls, n: int) -> VectorU8: ...
    def __init__(self, iterable: typing.Optional[typing.Collection[int]] = None): ...
    def __copy__(self) -> VectorU8: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> int: ...
    @typing.overload
    def __getitem__(self, index: slice) -> VectorU8: ...
    @typing.overload
    def __setitem__(self, index: int, value: int) -> None: ...
    @typing.overload
    def __setitem__(self, index: slice, value: int) -> None: ...
    def __add__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __iadd__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __sub__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __isub__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __mul__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __imul__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __floordiv__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __ifloordiv__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __matmul__(self, other: VectorU8) -> int: ...
    def argmax(self) -> int: ...
    def argmin(self) -> int: ...
    def copy(self) -> VectorU8: ...
    def max(self) -> int: ...
    def min(self) -> int: ...
    def reverse(self) -> None: ...
    def sum(self) -> int: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

class Matrix(typing.Sequence[Vector[_T]], typing.Generic[_T], abc.ABC):
    def __reduce__(self: _V) -> typing.Tuple[typing.Type[_V], typing.Tuple[typing.List[Vector[_T]]]]: ...
    @abc.abstractmethod
    @classmethod
    def zeros(cls: typing.Type[_M], m: int, n: int) -> _M: ...
    @abc.abstractmethod
    def __init__(self, iterable: typing.Collection[typing.Collection[_T]]): ...
    @abc.abstractmethod
    def __copy__(self: _M) -> _M: ...
    def __len__(self) -> int: ...
    @abc.abstractmethod
    @typing.overload
    def __getitem__(self, index: int) -> Vector[_T]: ...
    @abc.abstractmethod
    @typing.overload
    def __getitem__(self, index: slice) -> Matrix[_T]: ...
    @abc.abstractmethod
    @typing.overload
    def __getitem__(self, index: typing.Tuple[int, int]) -> _T: ...
    @abc.abstractmethod
    def __setitem__(self, index: typing.Tuple[int, int], value: _T) -> None: ...
    @abc.abstractmethod
    def __add__(self: _M, other: typing.Union[_T, _M]) -> _M: ...
    @abc.abstractmethod
    def __iadd__(self: _M, other: typing.Union[_T, _M]) -> _M: ...
    @abc.abstractmethod
    def __mul__(self: _M, other: typing.Union[_T, _M]) -> _M: ...
    @abc.abstractmethod
    def __imul__(self: _M, other: typing.Union[_T, _M]) -> _M: ...
    @abc.abstractmethod
    def __repr__(self) -> str: ...
    @property
    def shape(self) -> typing.Tuple[int, int]: ...
    @property
    def strides(self) -> typing.Tuple[int, int]: ...
    @property
    @abc.abstractmethod
    def itemsize(self) -> int: ...
    @property
    @abc.abstractmethod
    def format(self) -> str: ...
    @abc.abstractmethod
    def argmax(self) -> typing.Tuple[int, int]: ...
    @abc.abstractmethod
    def argmin(self) -> typing.Tuple[int, int]: ...
    @abc.abstractmethod
    def copy(self: _M) -> _M: ...
    @abc.abstractmethod
    def max(self) -> _T: ...
    @abc.abstractmethod
    def min(self) -> _T: ...
    @abc.abstractmethod
    def sum(self) -> _T: ...

class MatrixF(Matrix[float]):
    @classmethod
    def zeros(cls, m: int, n: int) -> MatrixF: ...
    def __init__(self, iterable: typing.Collection[typing.Collection[float]]): ...
    def __copy__(self) -> MatrixF: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> VectorF: ...
    @typing.overload
    def __getitem__(self, index: slice) -> MatrixF: ...
    @typing.overload
    def __getitem__(self, index: typing.Tuple[int, int]) -> float: ...
    def __setitem__(self, index: typing.Tuple[int, int], value: float) -> None: ...
    def __add__(self, other: typing.Union[float, MatrixF]) -> MatrixF: ...
    def __iadd__(self, other: typing.Union[float, MatrixF]) -> MatrixF: ...
    def __mul__(self, other: typing.Union[float, MatrixF]) -> MatrixF: ...
    def __imul__(self, other: typing.Union[float, MatrixF]) -> MatrixF: ...
    def __repr__(self) -> str: ...
    def __sizeof__(self) -> int: ...
    def argmax(self) -> typing.Tuple[int, int]: ...
    def argmin(self) -> typing.Tuple[int, int]: ...
    def copy(self) -> MatrixF: ...
    def max(self) -> float: ...
    def min(self) -> float: ...
    def sum(self) -> float: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

class MatrixU8(Matrix[int]):
    @classmethod
    def zeros(cls, m: int, n: int) -> MatrixU8: ...
    def __init__(self, iterable: typing.Collection[typing.Collection[float]]): ...
    def __copy__(self) -> MatrixU8: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> VectorU8: ...
    @typing.overload
    def __getitem__(self, index: slice) -> MatrixU8: ...
    @typing.overload
    def __getitem__(self, index: typing.Tuple[int, int]) -> int: ...
    def __setitem__(self, index: typing.Tuple[int, int], value: int) -> None: ...
    def __add__(self, other: typing.Union[int, MatrixU8]) -> MatrixU8: ...
    def __iadd__(self, other: typing.Union[int, MatrixU8]) -> MatrixU8: ...
    def __mul__(self, other: typing.Union[int, MatrixU8]) -> MatrixU8: ...
    def __imul__(self, other: typing.Union[int, MatrixU8]) -> MatrixU8: ...
    def __repr__(self) -> str: ...
    def __sizeof__(self) -> int: ...
    def argmax(self) -> typing.Tuple[int, int]: ...
    def argmin(self) -> typing.Tuple[int, int]: ...
    def copy(self) -> MatrixU8: ...
    def max(self) -> int: ...
    def min(self) -> int: ...
    def sum(self) -> int: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

# --- Multiple Sequences Alignment -------------------------------------------

class _MSASequences(typing.Sequence[Sequence], abc.ABC):
    def __len__(self) -> int: ...
    @abc.abstractmethod
    def __getitem__(self, idx: int) -> Sequence: ...  # type: ignore

class MSA(abc.ABC, typing.Sized):
    @abc.abstractmethod
    def __init__(
        self, nsequences: int, length: typing.Optional[int] = None
    ) -> None: ...
    def __copy__(self) -> MSA: ...
    def __eq__(self, other: object) -> bool: ...
    def __len__(self) -> int: ...
    @property
    def accession(self) -> typing.Optional[bytes]: ...
    @accession.setter
    def accession(self, accession: typing.Optional[bytes]) -> None: ...
    @property
    def author(self) -> typing.Optional[bytes]: ...
    @author.setter
    def author(self, author: typing.Optional[bytes]) -> None: ...
    @property
    def name(self) -> typing.Optional[bytes]: ...
    @name.setter
    def name(self, name: typing.Optional[bytes]) -> None: ...
    @property
    def description(self) -> typing.Optional[bytes]: ...
    @description.setter
    def description(self, description: typing.Optional[bytes]) -> None: ...
    @property
    def names(self) -> typing.Tuple[bytes]: ...
    @abc.abstractmethod
    def copy(self) -> MSA: ...
    def checksum(self) -> int: ...
    def write(self, fh: typing.BinaryIO, format: str) -> None: ...

class _TextMSASequences(_MSASequences, typing.Sequence[TextSequence]):
    def __init__(self, msa: TextMSA) -> None: ...
    def __getitem__(self, idx: int) -> TextSequence: ...  # type: ignore
    def __setitem__(self, idx: int, item: TextSequence) -> None: ...

class TextMSA(MSA):
    def __init__(
        self,
        name: typing.Optional[bytes] = None,
        description: typing.Optional[bytes] = None,
        accession: typing.Optional[bytes] = None,
        sequences: typing.Optional[typing.Iterable[TextSequence]] = None,
        author: typing.Optional[bytes] = None,
    ) -> None: ...
    def __copy__(self) -> TextMSA: ...
    def copy(self) -> TextMSA: ...
    def digitize(self, alphabet: Alphabet) -> DigitalMSA: ...
    @property
    def alignment(self) -> typing.Tuple[str]: ...
    @property
    def sequences(self) -> _TextMSASequences: ...

class _DigitalMSASequences(_MSASequences, typing.Sequence[DigitalSequence]):
    alphabet: Alphabet
    def __init__(self, msa: DigitalMSA) -> None: ...
    def __getitem__(self, idx: int) -> DigitalSequence: ...  # type: ignore
    def __setitem__(self, idx: int, item: DigitalSequence) -> None: ...

class DigitalMSA(MSA):
    alphabet: Alphabet
    def __init__(
        self,
        alphabet: Alphabet,
        name: typing.Optional[bytes] = None,
        description: typing.Optional[bytes] = None,
        accession: typing.Optional[bytes] = None,
        sequences: typing.Optional[typing.Iterable[DigitalSequence]] = None,
        author: typing.Optional[bytes] = None,
    ) -> None: ...
    def __copy__(self) -> DigitalMSA: ...
    def copy(self) -> DigitalMSA: ...
    def textize(self) -> TextMSA: ...
    @property
    def sequences(self) -> _DigitalMSASequences: ...

# --- MSA File ---------------------------------------------------------------

class MSAFile(typing.ContextManager[MSAFile], typing.Iterator[MSA]):
    _FORMATS: typing.ClassVar[typing.Dict[str, int]]
    alphabet: typing.Optional[Alphabet]
    def __init__(
        self,
        file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr], typing.BinaryIO],
        format: typing.Optional[str] = None
    ) -> None: ...
    def __enter__(self) -> MSAFile: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def __iter__(self) -> MSAFile: ...
    def __next__(self) -> MSA: ...
    @property
    def closed(self) -> bool: ...
    def read(self) -> typing.Optional[MSA]: ...
    def close(self) -> None: ...
    def guess_alphabet(self) -> typing.Optional[Alphabet]: ...
    def set_digital(self, alphabet: Alphabet) -> Alphabet: ...

# --- Randomness -------------------------------------------------------------

class Randomness(object):
    def __init__(
        self, seed: typing.Optional[int] = None, fast: bool = False
    ) -> None: ...
    def __copy__(self) -> Randomness: ...
    def __getstate__(self) -> typing.Tuple[typing.Any, ...]: ...
    def __setstate__(self, state: typing.Tuple[typing.Any, ...]) -> None: ...
    def __sizeof__(self) -> int: ...
    def getstate(self) -> typing.Tuple[typing.Any, ...]: ...
    def setstate(self, state: typing.Tuple[typing.Any, ...]) -> None: ...
    def seed(self, n: typing.Optional[int] = None) -> None: ...
    def copy(self) -> Randomness: ...
    def random(self) -> float: ...
    def normalvariate(self, mu: float, sigma: float) -> float: ...
    def is_fast(self) -> bool: ...

# --- Sequence ---------------------------------------------------------------

class Sequence(typing.Sized, abc.ABC):
    @abc.abstractmethod
    def __init__(self) -> None: ...
    def __copy__(self) -> Sequence: ...
    def __len__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    @property
    def accession(self) -> bytes: ...
    @accession.setter
    def accession(self, accession: bytes) -> None: ...
    @property
    def description(self) -> bytes: ...
    @description.setter
    def description(self, description: bytes) -> None: ...
    @property
    def name(self) -> bytes: ...
    @name.setter
    def name(self, name: bytes) -> None: ...
    @property
    def source(self) -> bytes: ...
    @source.setter
    def source(self, src: bytes) -> None: ...
    @property
    def taxonomy_id(self) -> typing.Optional[int]: ...
    @taxonomy_id.setter
    def taxonomy_id(self, tax_id: typing.Optional[int]) -> None: ...
    @property
    def residue_markups(self) -> typing.Dict[bytes, bytes]: ...
    @residue_markups.setter
    def residue_markups(self, xr: typing.Dict[bytes, bytes]) -> None: ...
    def checksum(self) -> int: ...
    def clear(self) -> None: ...
    @abc.abstractmethod
    def copy(self) -> Sequence: ...
    def write(self, fh: typing.BinaryIO) -> None: ...
    if sys.version_info >= (3, 8):
        @typing.overload
        def reverse_complement(self, inplace: typing.Literal[True]) -> None: ...
        @typing.overload
        def reverse_complement(
            self, inplace: typing.Literal[False] = False
        ) -> Sequence: ...
    else:
        @typing.overload
        def reverse_complement(
            self, inplace: bool = False
        ) -> typing.Optional[Sequence]: ...

class TextSequence(Sequence):
    def __init__(
        self,
        name: bytes = None,
        description: bytes = None,
        accession: bytes = None,
        sequence: str = None,
        source: bytes = None,
    ) -> None: ...
    def copy(self) -> TextSequence: ...
    def digitize(self, alphabet: Alphabet) -> DigitalSequence: ...
    @property
    def sequence(self) -> str: ...
    if sys.version_info >= (3, 8):
        @typing.overload
        def reverse_complement(self, inplace: typing.Literal[True]) -> None: ...
        @typing.overload
        def reverse_complement(
            self, inplace: typing.Literal[False] = False
        ) -> TextSequence: ...
    else:
        def reverse_complement(
            self, inplace: bool = False
        ) -> typing.Optional[TextSequence]: ...

class DigitalSequence(Sequence):
    alphabet: Alphabet
    def __init__(
        self,
        alphabet: Alphabet,
        name: bytes = None,
        description: bytes = None,
        accession: bytes = None,
        sequence: str = None,
        source: bytes = None,
    ) -> None: ...
    def copy(self) -> DigitalSequence: ...
    def textize(self) -> TextSequence: ...
    @property
    def sequence(self) -> VectorU8: ...
    if sys.version_info >= (3, 8):
        @typing.overload
        def reverse_complement(self, inplace: typing.Literal[True]) -> None: ...
        @typing.overload
        def reverse_complement(
            self, inplace: typing.Literal[False] = False
        ) -> DigitalSequence: ...
    else:
        def reverse_complement(
            self, inplace: bool = False
        ) -> typing.Optional[DigitalSequence]: ...

# --- Sequence File ----------------------------------------------------------

class SequenceFile(typing.ContextManager[SequenceFile], typing.Iterator[Sequence]):
    _FORMATS: typing.ClassVar[typing.Dict[str, int]]
    alphabet: typing.Optional[Alphabet]
    @classmethod
    def parse(cls, buffer: bytes, format: str) -> Sequence: ...
    @classmethod
    def parseinto(cls, seq: Sequence, buffer: bytes, format: str) -> Sequence: ...
    def __init__(
        self,
        file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr], typing.BinaryIO],
        format: typing.Optional[str] = None,
        ignore_gaps: bool = False
    ) -> None: ...
    def __enter__(self) -> SequenceFile: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def __iter__(self) -> SequenceFile: ...
    def __next__(self) -> Sequence: ...
    @property
    def closed(self) -> bool: ...
    def read(self) -> typing.Optional[Sequence]: ...
    def read_info(self) -> typing.Optional[Sequence]: ...
    def read_seq(self) -> typing.Optional[Sequence]: ...
    def readinto(self, seq: Sequence) -> typing.Optional[Sequence]: ...
    def readinto_seq(self, seq: Sequence) -> typing.Optional[Sequence]: ...
    def readinto_info(self, seq: Sequence) -> typing.Optional[Sequence]: ...
    def close(self) -> None: ...
    def guess_alphabet(self) -> typing.Optional[Alphabet]: ...
    def set_digital(self, alphabet: Alphabet) -> Alphabet: ...

# --- Sequence/Subsequence Index ---------------------------------------------

class SSIReader(object):

    class Entry(typing.NamedTuple):
        fd: int
        record_offset: int
        data_offset: int
        record_length: int

    class FileInfo(typing.NamedTuple):
        name: str
        format: int

    def __init__(
        self, file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr]]
    ) -> None: ...
    def __enter__(self) -> SSIReader: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def close(self) -> None: ...

class SSIWriter(object):
    def __init__(
        self, file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr]]
    ) -> None: ...
    def __enter__(self) -> SSIWriter: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def add_alias(self, alias: bytes, key: bytes) -> None: ...
    def add_file(self, filename: str, format: int = 0) -> int: ...
    def add_key(
        self,
        key: bytes,
        fd: int,
        record_offset: int,
        data_offset: int = 0,
        record_length: int = 0,
    ) -> None: ...
    def close(self) -> None: ...



# coding: utf-8
"""Obtain a `FILE*` from a Python object using ``funopen``.
"""

# --- C imports --------------------------------------------------------------

from cpython.buffer cimport Py_buffer, PyObject_GetBuffer, PyBuffer_Release, PyBUF_READ, PyBUF_WRITE
from cpython.memoryview cimport PyMemoryView_FromMemory
from cpython.ref cimport PyObject, Py_INCREF, Py_DECREF
from libc.stdio cimport EOF, FILE
from libc.stdint cimport int64_t, uint64_t
from libc.string cimport strcpy, strncpy, memcpy


# --- BSD interface ----------------------------------------------------------

cdef extern from "stdio.h":
    ctypedef long fpos_t

ctypedef int    (*readfn_t) (void *cookie,       char *buf, int size)
ctypedef int    (*writefn_t)(void *cookie, const char *buf, int size)
ctypedef fpos_t (*seekfn_t) (void *cookie,   fpos_t offset, int whence)
ctypedef int    (*closefn_t)(void *cookie)

cdef extern from "stdio.h":
    FILE* funopen(
        const void* cookie,
        readfn_t    readfn,
        writefn_t   writefn,
        seekfn_t    seekfn,
        closefn_t   closefn
    )


# --- fwrite implementation --------------------------------------------------

cdef int fwrite_obj(void *cookie, const char *buf, int size) except 0:
    """Zero-copy implementation of `fwrite` for a Python file-handle.
    """
    cdef object obj = <object> cookie
    cdef object mem = PyMemoryView_FromMemory(<char*> buf, size, PyBUF_READ)
    return obj.write(mem)


# --- fread implementations --------------------------------------------------

cdef int fread_obj_read(void *cookie, char *buf, int size) except -1:
    """Copying variant of `fread` for files lacking `readinto`.
    """
    cdef object    obj      = <object> cookie
    cdef object    chunk    = obj.read(size)
    cdef Py_buffer pybuffer

    if PyObject_GetBuffer(chunk, &pybuffer, PyBUF_READ) < 0:
        raise RuntimeError("could not get buffer")
    memcpy(buf, pybuffer.buf, len(chunk))
    PyBuffer_Release(&pybuffer)

    return len(chunk)


cdef int fread_obj_readinto(void *cookie, char *buf, int size) except -1:
    """Zero-copy implementation of `fread` using the `readinto` method.
    """
    cdef object obj   = <object> cookie
    cdef object mem

    IF SYS_IMPLEMENTATION_NAME == "pypy":
        # NB: PyPy has a bug in the `readinto` implementation that requires the
        #     memoryview to be read/write and not just write, which is why we
        #     create the memoryview in read/write mode and not just in write mode.
        mem = PyMemoryView_FromMemory(buf, size, PyBUF_READ | PyBUF_WRITE)
    ELSE:
        mem = PyMemoryView_FromMemory(buf, size, PyBUF_WRITE)

    # NB: Make sure to cast the memoryview as unsigned char* : on PyPy,
    #     `obj.readinto` sometimes complains that the memoryview is not
    #     a byte-like object.
    return obj.readinto(mem.cast('B'))


# --- fseek implementation ---------------------------------------------------

cdef fpos_t fseek_obj(void* cookie, fpos_t offset, int whence) except -1:
    cdef object obj = <object> cookie
    return obj.seek(offset, whence)


# --- fclose implementation --------------------------------------------------

cdef int fclose_obj(void *cookie) except EOF:
    Py_DECREF(<object> cookie)
    return 0


# --- fopen_obj --------------------------------------------------------------

cdef FILE* fopen_obj(object obj, str mode = "r") except NULL:
    cdef closefn_t closefn = <closefn_t> fclose_obj
    cdef readfn_t  readfn  = NULL
    cdef writefn_t writefn = NULL
    cdef seekfn_t  seekfn  = NULL

    try:
        if obj.readable():
            if hasattr((<object> obj), "readinto"):
                readfn = <readfn_t> fread_obj_readinto
            else:
                readfn = <readfn_t> fread_obj_read
        if obj.writable():
            writefn = <writefn_t> fwrite_obj
        if obj.seekable():
            seekfn = <seekfn_t> fseek_obj
    except AttributeError as err:
        ty = type(obj).__name__
        raise TypeError("expected `io.IOBase` instance, found {}".format(ty)) from err

    Py_INCREF(obj)
    return funopen(<void*> obj, readfn, writefn, seekfn, closefn)



from libc.stdio cimport FILE

cdef extern from "esl_msafile_clustal.h" nogil:
    int esl_msafile_clustal_SetInmap     (ESL_MSAFILE *afp)
    int esl_msafile_clustal_GuessAlphabet(ESL_MSAFILE *afp, int *ret_type)
    int esl_msafile_clustal_Read         (ESL_MSAFILE *afp, ESL_MSA **ret_msa)
    int esl_msafile_clustal_Write        (FILE *fp,    const ESL_MSA *msa, int fmt)



# coding: utf-8
"""Reimplementation of HMMER binaries with the pyHMMER API.
"""

import abc
import contextlib
import collections
import ctypes
import itertools
import io
import queue
import time
import threading
import typing
import os
import multiprocessing

import psutil

from .easel import Alphabet, DigitalSequence, DigitalMSA, MSA, MSAFile, TextSequence, SequenceFile, SSIWriter
from .plan7 import Builder, Background, Pipeline, LongTargetsPipeline, TopHits, HMM, HMMFile, Profile, TraceAligner, OptimizedProfile
from .utils import peekable

# the query type for the pipeline
_Q = typing.TypeVar("_Q", HMM, Profile, OptimizedProfile, DigitalSequence, DigitalMSA)
# the model type for the pipeline
_M = typing.TypeVar("_M", HMM, Profile, OptimizedProfile)
# the sequence type for the pipeline
_S = typing.TypeVar("_S", DigitalSequence, DigitalMSA)

# --- Pipeline threads -------------------------------------------------------

class _PipelineThread(typing.Generic[_Q], threading.Thread):
    """A generic worker thread to parallelize a pipelined search.

    Attributes:
        sequence (iterable of `DigitalSequence`): The target sequences to
            search for hits. **Must be able to be iterated upon more than
            once.**
        query_queue (`queue.Queue`): The queue used to pass queries between
            threads. It contains both the query and its index, so that the
            results can be returned in the same order.
        query_count (`multiprocessing.Value`): An atomic counter storing
            the total number of queries that have currently been loaded.
            Passed to the ``callback`` so that an UI can show the total
            for a progress bar.
        hits_queue (`queue.PriorityQueue`): The queue used to pass back
            the `TopHits` to the main thread. The results are inserted
            using the index of the query, so that the main thread can
            pull results in order.
        kill_switch (`threading.Event`): An event flag shared between
            all worker threads, used to notify emergency exit.
        hits_found (`list` of `threading.Event`): A list of event flags,
            such that ``hits_found[i]`` is set when results have been
            obtained for the query of index ``i``. This allows the main
            thread to keep waiting for the right `TopHits` to yield even
            if subsequent queries have already been treated, and to make
            sure the next result returned by ``hits_queue.get`` will also
            be of index ``i``.
        callback (`callable`, optional): An optional callback to be called
            after each query has been processed. It should accept two
            arguments: the query object that was processed, and the total
            number of queries read until now.
        options (`dict`): A dictionary of options to be passed to the
            `pyhmmer.plan7.Pipeline` object wrapped by the worker thread.
        pipeline_class (`type`): The pipeline class to use to search for
            hits. Use `~plan7.LongTargetsPipeline` for `nhmmer`, and
            `~plan7.Pipeline` everywhere else.

    """

    @staticmethod
    def _none_callback(hmm: _Q, total: int) -> None:
        pass

    def __init__(
        self,
        sequences: typing.Iterable[DigitalSequence],
        query_queue: "queue.Queue[typing.Optional[typing.Tuple[int, _Q]]]",
        query_count: multiprocessing.Value,  # type: ignore
        hits_queue: "queue.PriorityQueue[typing.Tuple[int, TopHits]]",
        kill_switch: threading.Event,
        hits_found: typing.List[threading.Event],
        callback: typing.Optional[typing.Callable[[_Q, int], None]],
        options: typing.Dict[str, typing.Any],
        pipeline_class: typing.Type[Pipeline],
        alphabet: Alphabet,
    ) -> None:
        super().__init__()
        self.options = options
        self.sequences = list(sequences)
        self.pipeline = pipeline_class(alphabet=alphabet, **options)
        self.query_queue: "queue.Queue[typing.Optional[typing.Tuple[int, _Q]]]" = query_queue
        self.query_count = query_count
        self.hits_queue = hits_queue
        self.callback: "typing.Optional[typing.Callable[[_Q, int], None]]" = callback or self._none_callback
        self.kill_switch = kill_switch
        self.hits_found = hits_found
        self.error: typing.Optional[BaseException] = None

    def run(self) -> None:
        while not self.kill_switch.is_set():
            # attempt to get the next argument, with a timeout
            # so that the thread can periodically check if it has
            # been killed, even when the query queue is empty
            try:
                args = self.query_queue.get(timeout=1)
            except queue.Empty:
                continue
            # check if arguments from the queue are a poison-pill (`None`),
            # in which case the thread will stop running
            if args is None:
                self.query_queue.task_done()
                return
            else:
                index, query = args
            # process the arguments, making sure to capture any exception
            # raised while processing the query, and then mark the hits
            # as "found" using a `threading.Event` for each.
            try:
                self.process(index, query)
                self.query_queue.task_done()
            except BaseException as exc:
                self.error = exc
                self.kill()
                return
            finally:
                self.hits_found[index].set()

    def kill(self) -> None:
        self.kill_switch.set()

    def process(self, index: int, query: _Q) -> None:
        hits = self.search(query)
        self.hits_queue.put((index, hits))
        self.callback(query, self.query_count.value)  # type: ignore
        self.pipeline.clear()

    @abc.abstractmethod
    def search(self, query: _Q) -> TopHits:
        return NotImplemented


class _ModelPipelineThread(typing.Generic[_M], _PipelineThread[_M]):
    def search(self, query: _M) -> TopHits:
        return self.pipeline.search_hmm(query, self.sequences)


class _SequencePipelineThread(_PipelineThread[DigitalSequence]):
    def __init__(
        self,
        sequences: typing.Iterable[DigitalSequence],
        query_queue: "queue.Queue[typing.Optional[typing.Tuple[int, DigitalSequence]]]",
        query_count: multiprocessing.Value,  # type: ignore
        hits_queue: "queue.PriorityQueue[typing.Tuple[int, TopHits]]",
        kill_switch: threading.Event,
        hits_found: typing.List[threading.Event],
        callback: typing.Optional[typing.Callable[[DigitalSequence, int], None]],
        options: typing.Dict[str, typing.Any],
        pipeline_class: typing.Type[Pipeline],
        alphabet: Alphabet,
        builder: Builder,
    ) -> None:
        super().__init__(
            sequences,
            query_queue,
            query_count,
            hits_queue,
            kill_switch,
            hits_found,
            callback,
            options,
            pipeline_class,
            alphabet,
        )
        self.builder = builder

    def search(self, query: DigitalSequence) -> TopHits:
        return self.pipeline.search_seq(query, self.sequences, self.builder)


class _MSAPipelineThread(_PipelineThread[DigitalMSA]):
    def __init__(
        self,
        sequences: typing.Iterable[DigitalSequence],
        query_queue: "queue.Queue[typing.Optional[typing.Tuple[int, DigitalMSA]]]",
        query_count: multiprocessing.Value,  # type: ignore
        hits_queue: "queue.PriorityQueue[typing.Tuple[int, TopHits]]",
        kill_switch: threading.Event,
        hits_found: typing.List[threading.Event],
        callback: typing.Optional[typing.Callable[[DigitalMSA, int], None]],
        options: typing.Dict[str, typing.Any],
        pipeline_class: typing.Type[Pipeline],
        alphabet: Alphabet,
        builder: Builder,
    ) -> None:
        super().__init__(
            sequences,
            query_queue,
            query_count,
            hits_queue,
            kill_switch,
            hits_found,
            callback,
            options,
            pipeline_class,
            alphabet,
        )
        self.builder = builder

    def search(self, query: DigitalMSA) -> TopHits:
        return self.pipeline.search_msa(query, self.sequences, self.builder)


# --- Search runners ---------------------------------------------------------

class _Search(typing.Generic[_Q], abc.ABC):

    def __init__(
        self,
        queries: typing.Iterable[_Q],
        sequences: typing.Collection[DigitalSequence],
        cpus: int = 0,
        callback: typing.Optional[typing.Callable[[_Q, int], None]] = None,
        pipeline_class: typing.Type[Pipeline] = Pipeline,
        alphabet: Alphabet = Alphabet.amino(),
        **options: typing.Dict[str, object],
    ) -> None:
        self.queries: typing.Iterable[_Q] = queries
        self.sequences = sequences
        self.cpus = cpus
        self.callback: typing.Optional[typing.Callable[[_Q, int], None]] = callback
        self.options = options
        self.pipeline_class = pipeline_class
        self.alphabet = alphabet

    @abc.abstractmethod
    def _new_thread(
        self,
        query_queue: "queue.Queue[typing.Optional[typing.Tuple[int, _Q]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        hits_queue: "queue.PriorityQueue[typing.Tuple[int, TopHits]]",
        kill_switch: threading.Event,
        hits_found: typing.List[threading.Event],
    ) -> _PipelineThread[_Q]:
        return NotImplemented

    def _single_threaded(self) -> typing.Iterator[TopHits]:
        # create the queues to pass the HMM objects around, as well as atomic
        # values that we use to synchronize the threads
        hits_found: typing.List[threading.Event] = []
        query_queue = queue.Queue()  # type: ignore
        query_count = multiprocessing.Value(ctypes.c_ulong)
        hits_queue = queue.PriorityQueue()  # type: ignore
        kill_switch = threading.Event()

        # create the thread (to recycle code)
        thread = self._new_thread(query_queue, query_count, hits_queue, kill_switch, hits_found)

        # process each HMM iteratively and yield the result
        # immediately so that the user can iterate over the
        # TopHits one at a time
        for index, query in enumerate(self.queries):
            query_count.value += 1
            thread.process(index, query)
            yield hits_queue.get_nowait()[1]

    def _multi_threaded(self) -> typing.Iterator[TopHits]:
        # create the queues to pass the HMM objects around, as well as atomic
        # values that we use to synchronize the threads
        hits_found: typing.List[threading.Event] = []
        hits_queue = queue.PriorityQueue()  # type: ignore
        query_count = multiprocessing.Value(ctypes.c_ulong)
        kill_switch = threading.Event()
        # the query queue is bounded so that we only feed more queries
        # if the worker threads are waiting for some
        query_queue = queue.Queue(maxsize=self.cpus)  # type: ignore
        # additional type annotations
        query: typing.Optional[_Q]
        index: int

        # create and launch one pipeline thread per CPU
        threads = []
        for _ in range(self.cpus):
            thread = self._new_thread(query_queue, query_count, hits_queue, kill_switch, hits_found)
            thread.start()
            threads.append(thread)

        # catch exceptions to kill threads in the background before exiting
        try:
            # enumerate queries, so that we now the index of each query
            # and we can yield the results in the same order
            queries = enumerate(self.queries)
            # initially feed one query per thread so that they can start
            # working before we enter the main loop
            for (index, query) in itertools.islice(queries, self.cpus):
                query_count.value += 1
                hits_found.append(threading.Event())
                query_queue.put((index, query))
            # alternate between feeding queries to the threads and
            # yielding back results, if available
            hits_yielded = 0
            while hits_yielded < query_count.value:
                # get the next query, or break the loop if there is no query
                # left to process in the input iterator.
                index, query = next(queries, (-1, None))
                if query is None:
                    break
                else:
                    query_count.value += 1
                    hits_found.append(threading.Event())
                    query_queue.put((index, query))
                # yield the top hits for the next query, if available
                if hits_found[hits_yielded].is_set():
                    yield hits_queue.get_nowait()[1]
                    hits_yielded += 1
            # now that we exhausted all queries, poison pill the
            # threads so they stop on their own
            for _ in threads:
                query_queue.put(None)
            # yield remaining results
            while hits_yielded < query_count.value:
                hits_found[hits_yielded].wait()
                yield hits_queue.get_nowait()[1]
                hits_yielded += 1
        except queue.Empty:
            # the only way we can get queue.Empty is if a thread has set
            # the flag for `hits_found[i]` without actually putting it in
            # the queue: this only happens when a background thread raised
            # an exception, so we must chain it
            for thread in threads:
                if thread.error is not None:
                    raise thread.error from None
            # if this is exception is otherwise a bug, make sure to reraise it
            raise
        except BaseException:
            # make sure threads are killed to avoid being stuck,
            # e.g. after a KeyboardInterrupt
            kill_switch.set()
            raise

    def run(self) -> typing.Iterator[TopHits]:
        if self.cpus == 1:
            return self._single_threaded()
        else:
            return self._multi_threaded()


class _ModelSearch(typing.Generic[_M], _Search[_M]):

    def _new_thread(
        self,
        query_queue: "queue.Queue[typing.Optional[typing.Tuple[int, _M]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        hits_queue: "queue.PriorityQueue[typing.Tuple[int, TopHits]]",
        kill_switch: threading.Event,
        hits_found: typing.List[threading.Event],
    ) -> _ModelPipelineThread[_M]:
        return _ModelPipelineThread(
            self.sequences,
            query_queue,
            query_count,
            hits_queue,
            kill_switch,
            hits_found,
            self.callback,
            self.options,
            self.pipeline_class,
            self.alphabet,
        )


class _SequenceSearch(_Search[DigitalSequence]):

    def __init__(
        self,
        builder: Builder,
        queries: typing.Iterable[DigitalSequence],
        sequences: typing.Collection[DigitalSequence],
        cpus: int = 0,
        callback: typing.Optional[typing.Callable[[DigitalSequence, int], None]] = None,
        pipeline_class: typing.Type[Pipeline] = Pipeline,
        alphabet: Alphabet = Alphabet.amino(),
        **options, # type: typing.Dict[str, object],
    ) -> None:
        super().__init__(queries, sequences, cpus, callback, pipeline_class, alphabet, **options)
        self.builder = builder

    def _new_thread(
        self,
        query_queue: "queue.Queue[typing.Optional[typing.Tuple[int, DigitalSequence]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        hits_queue: "queue.PriorityQueue[typing.Tuple[int, TopHits]]",
        kill_switch: threading.Event,
        hits_found: typing.List[threading.Event],
    ) -> _SequencePipelineThread:
        return _SequencePipelineThread(
            self.sequences,
            query_queue,
            query_count,
            hits_queue,
            kill_switch,
            hits_found,
            self.callback,
            self.options,
            self.pipeline_class,
            self.alphabet,
            self.builder.copy(),
        )


class _MSASearch(_Search[DigitalMSA]):

    def __init__(
        self,
        builder: Builder,
        queries: typing.Iterable[DigitalMSA],
        sequences: typing.Collection[DigitalSequence],
        cpus: int = 0,
        callback: typing.Optional[typing.Callable[[DigitalMSA, int], None]] = None,
        pipeline_class: typing.Type[Pipeline] = Pipeline,
        alphabet: Alphabet = Alphabet.amino(),
        **options, # type: typing.Dict[str, object],
    ) -> None:
        super().__init__(queries, sequences, cpus, callback, pipeline_class, alphabet, **options)
        self.builder = builder

    def _new_thread(
        self,
        query_queue: "queue.Queue[typing.Optional[typing.Tuple[int, DigitalMSA]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        hits_queue: "queue.PriorityQueue[typing.Tuple[int, TopHits]]",
        kill_switch: threading.Event,
        hits_found: typing.List[threading.Event],
    ) -> _MSAPipelineThread:
        return _MSAPipelineThread(
            self.sequences,
            query_queue,
            query_count,
            hits_queue,
            kill_switch,
            hits_found,
            self.callback,
            self.options,
            self.pipeline_class,
            self.alphabet,
            self.builder.copy(),
        )


# --- hmmsearch --------------------------------------------------------------

def hmmsearch(
    queries: typing.Iterable[_M],
    sequences: typing.Collection[DigitalSequence],
    cpus: int = 0,
    callback: typing.Optional[typing.Callable[[_M, int], None]] = None,
    **options,  # type: typing.Dict[str, object]
) -> typing.Iterator[TopHits]:
    """Search HMM profiles against a sequence database.

    Arguments:
        queries (iterable of `HMM`, `Profile` or `OptimizedProfile`): The
            query HMMs or profiles to search for in the database.
        sequences (collection of `~pyhmmer.easel.DigitalSequence`): A
            database of sequences to query.
        cpus (`int`): The number of threads to run in parallel. Pass ``1``
            to run everything in the main thread, ``0`` to automatically
            select a suitable number (using `psutil.cpu_count`), or any
            positive number otherwise.
        callback (callable): A callback that is called everytime a query is
            processed with two arguments: the query, and the total number
            of queries. This can be used to display progress in UI.

    Yields:
        `~pyhmmer.plan7.TopHits`: An object reporting *top hits* for each
        query, in the same order the queries were passed in the input.

    Raises:
        `~pyhmmer.errors.AlphabetMismatch`: When any of the query HMMs
        and the sequences do not share the same alphabet.

    Note:
        Any additional arguments passed to the `hmmsearch` function will be
        passed transparently to the `~pyhmmer.plan7.Pipeline` to be created.

    .. versionadded:: 0.1.0

    .. versionchanged:: 0.4.9
       Allow using `Profile` and `OptimizedProfile` queries.

    """
    # count the number of CPUs to use
    _cpus = cpus if cpus > 0 else psutil.cpu_count(logical=False) or multiprocessing.cpu_count()
    runner: _ModelSearch[_M] = _ModelSearch(queries, sequences, _cpus, callback, **options) # type: ignore
    return runner.run()


# --- phmmer -----------------------------------------------------------------

def phmmer(
    queries: typing.Iterable[_S],
    sequences: typing.Collection[DigitalSequence],
    cpus: int = 0,
    callback: typing.Optional[typing.Callable[[_S, int], None]] = None,
    builder: typing.Optional[Builder] = None,
    **options, # type: typing.Dict[str, object]
) -> typing.Iterator[TopHits]:
    """Search protein sequences against a sequence database.

    Arguments:
        queries (iterable of `DigitalSequence` or `DigitalMSA`): The
            query sequences to search for in the sequence database.
        sequences (collection of `~pyhmmer.easel.DigitalSequence`): A
            database of sequences to query.
        cpus (`int`): The number of threads to run in parallel. Pass ``1`` to
            run everything in the main thread, ``0`` to automatically
            select a suitable number (using `psutil.cpu_count`), or any
            positive number otherwise.
        callback (callable): A callback that is called everytime a query is
            processed with two arguments: the query, and the total number
            of queries. This can be used to display progress in UI.
        builder (`~pyhmmer.plan7.Builder`, optional): A builder to configure
            how the queries are converted to HMMs. Passing `None` will create
            a default instance.

    Yields:
        `~pyhmmer.plan7.TopHits`: A *top hits* instance for each query,
        in the same order the queries were passed in the input.

    Note:
        Any additional keyword arguments passed to the `phmmer` function
        will be passed transparently to the `~pyhmmer.plan7.Pipeline` to
        be created in each worker thread.

    .. versionadded:: 0.2.0

    .. versionchanged:: 0.3.0
       Allow using `DigitalMSA` queries.

    """
    _cpus = cpus if cpus > 0 else psutil.cpu_count(logical=False) or multiprocessing.cpu_count()
    _builder = Builder(Alphabet.amino()) if builder is None else builder

    try:
        _queries: peekable[typing.Union[DigitalSequence, DigitalMSA, HMM]] = peekable(queries)
        _item: typing.Union[DigitalSequence, DigitalMSA, HMM, None] = _queries.peek()
    except StopIteration:
        _item = None

    runner: typing.Union[_SequenceSearch, _MSASearch]
    if _item is None or isinstance(_item, DigitalSequence):
        runner = _SequenceSearch(
            _builder,
            typing.cast(peekable[DigitalSequence], _queries),
            sequences,
            _cpus,
            callback,  # type: ignore
            pipeline_class=Pipeline,
            alphabet=Alphabet.amino(),
            **options
        )
    elif isinstance(_item, DigitalMSA):
        runner = _MSASearch(
            _builder, _queries, sequences, _cpus, callback, pipeline_class=Pipeline, alphabet=Alphabet.amino(), **options   # type: ignore
        )
    else:
        name = type(_item).__name__
        raise TypeError(f"Expected iterable of DigitalSequence or DigitalMSA, found {name}")

    return runner.run()


# --- nhmmer -----------------------------------------------------------------

def nhmmer(
    queries: typing.Iterable[_Q],
    sequences: typing.Collection[DigitalSequence],
    cpus: int = 0,
    callback: typing.Optional[typing.Callable[[_Q, int], None]] = None,
    builder: typing.Optional[Builder] = None,
    **options, # type: typing.Dict[str, object]
) -> typing.Iterator[TopHits]:
    """Search nucleotide sequences against a sequence database.

    Arguments:
        queries (iterable of `DigitalSequence`, `DigitalMSA`, `HMM`): The
            query sequences or profiles to search for in the sequence
            database.
        sequences (collection of `~pyhmmer.easel.DigitalSequence`): A
            database of sequences to query.
        cpus (`int`): The number of threads to run in parallel. Pass ``1`` to
            run everything in the main thread, ``0`` to automatically
            select a suitable number (using `psutil.cpu_count`), or any
            positive number otherwise.
        callback (callable): A callback that is called everytime a query is
            processed with two arguments: the query, and the total number
            of queries. This can be used to display progress in UI.
        builder (`~pyhmmer.plan7.Builder`, optional): A builder to configure
            how the queries are converted to HMMs. Passing `None` will create
            a default instance.

    Yields:
        `~pyhmmer.plan7.TopHits`: A *top hits* instance for each query,
        in the same order the queries were passed in the input.

    Note:
        Any additional keyword arguments passed to the `nhmmer` function
        will be passed to the `~pyhmmer.plan7.LongTargetsPipeline` created
        in each worker thread. The ``strand`` argument can be used to
        restrict the search on the direct or reverse strand.

    Hint:
        This function is not just `phmmer` for nucleotide sequences; it
        actually uses a `~pyhmmer.plan7.LongTargetsPipeline` internally
        instead of processing each target sequence in its entirety when
        searching for hits. This avoids hitting the maximum target size
        that can be used (100,000 residues), which may be a problem for
        some larger genomes.

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.4.9
       Allow using `Profile` and `OptimizedProfile` queries.

    """
    _cpus = cpus if cpus > 0 else psutil.cpu_count(logical=False) or multiprocessing.cpu_count()
    _builder = Builder(Alphabet.dna()) if builder is None else builder

    try:
        _queries: peekable[_Q] = peekable(queries)
        _item: typing.Optional[_Q] = _queries.peek()
    except StopIteration:
        _item = None

    runner: typing.Union[_SequenceSearch, _MSASearch, _ModelSearch[HMM]]
    if _item is None or isinstance(_item, DigitalSequence):
        runner = _SequenceSearch(
            _builder,
            typing.cast(peekable[DigitalSequence], _queries),
            sequences,
            _cpus,
            callback,  # type: ignore
            pipeline_class=LongTargetsPipeline,
            alphabet=_item.alphabet if _item is not None else Alphabet.dna(),  # type: ignore
            **options,
        )
    elif isinstance(_item, DigitalMSA):
        runner = _MSASearch(
            _builder,
            typing.cast(peekable[DigitalMSA], _queries),
            sequences,
            _cpus,
            callback,
            pipeline_class=LongTargetsPipeline,
            alphabet=_item.alphabet,
            **options,
        )
    elif isinstance(_item, (HMM, Profile, OptimizedProfile)):
        runner = _ModelSearch(
            typing.cast(peekable[HMM], _queries),
            sequences,
            _cpus,
            callback,  # type: ignore
            pipeline_class=LongTargetsPipeline,
            alphabet=_item.alphabet,
            **options,
        )
    else:
        name = type(_item).__name__
        raise TypeError(f"Expected iterable of DigitalSequence, DigitalMSA, HMM, Profile or OptimizedProfile, found {name}")
    return runner.run()


# --- hmmpress ---------------------------------------------------------------

def hmmpress(
    hmms: typing.Iterable[HMM], output: typing.Union[str, "os.PathLike[str]"],
) -> int:
    """Press several HMMs into a database.

    Calling this function will create 4 files at the given location:
    ``{output}.h3p`` (containing the optimized profiles),
    ``{output}.h3m`` (containing the binary HMMs),
    ``{output}.h3f`` (containing the MSV parameters), and
    ``{output}.h3i`` (the SSI index mapping the previous files).

    Arguments:
        hmms (iterable of `~pyhmmer.plan7.HMM`): The HMMs to be pressed
            together in the file.
        output (`str` or `os.PathLike`): The path to an output location
            where to write the different files.

    """
    DEFAULT_L = 400
    path = os.fspath(output)
    nmodel = 0

    with contextlib.ExitStack() as ctx:
        h3p = ctx.enter_context(open("{}.h3p".format(path), "wb"))
        h3m = ctx.enter_context(open("{}.h3m".format(path), "wb"))
        h3f = ctx.enter_context(open("{}.h3f".format(path), "wb"))
        h3i = ctx.enter_context(SSIWriter("{}.h3i".format(path)))
        fh = h3i.add_file(path, format=0)

        for hmm in hmms:
            # create the background model on the first iteration
            if nmodel == 0:
                bg = Background(hmm.alphabet)
                bg.L = DEFAULT_L

            # build the optimized models
            gm = Profile(hmm.M, hmm.alphabet)
            gm.configure(hmm, bg, DEFAULT_L)
            om = gm.optimized()

            # update the disk offsets of the optimized model to be written
            om.offsets.model = h3m.tell()
            om.offsets.profile = h3p.tell()
            om.offsets.filter = h3f.tell()

            # check that hmm has a name
            if hmm.name is None:
                raise ValueError("HMMs must have a name to be pressed.")
            # add the HMM name, and optionally the HMM accession to the index
            h3i.add_key(hmm.name, fh, om.offsets.model, 0, 0)
            if hmm.accession is not None:
                h3i.add_alias(hmm.accession, hmm.name)

            # write the HMM in binary format, and the optimized profile
            hmm.write(h3m, binary=True)
            om.write(h3f, h3p)
            nmodel += 1

    # return the number of written HMMs
    return nmodel


# --- hmmalign ---------------------------------------------------------------

def hmmalign(
    hmm: HMM,
    sequences: typing.Collection[DigitalSequence],
    trim: bool = False,
    digitize: bool = False,
    all_consensus_cols: bool = True,
) -> MSA:
    """Align several sequences to a reference HMM, and return the MSA.

    Arguments:
        hmm (`~pyhmmer.plan7.HMM`): The reference HMM to use for the
            alignment.
        sequences (collection of `~pyhmmer.easel.DigitalSequence`): The
            sequences to align to the HMM.
        trim (`bool`): Trim off any residues that get assigned to
            flanking :math:`N` and :math:`C` states (in profile traces)
            or :math:`I_0` and :math:`I_m` (in core traces).
        digitize (`bool`): If set to `True`, returns a `DigitalMSA`
            instead of a `TextMSA`.
        all_consensus_cols (`bool`): Force a column to be created for
            every consensus column in the model, even if it means having
            all gap character in a column.

    Returns:
        `~pyhmmer.easel.MSA`: A multiple sequence alignment containing
        the aligned sequences, either a `TextMSA` or a `DigitalMSA`
        depending on the value of the ``digitize`` argument.

    See Also:
        The `~pyhmmer.plan7.TraceAligner` class, which lets you inspect the
        intermediate tracebacks obtained for each alignment before building
        a MSA.

    .. versionadded:: 0.4.7

    """
    aligner = TraceAligner()
    traces = aligner.compute_traces(hmm, sequences)
    return aligner.align_traces(
        hmm,
        sequences,
        traces,
        trim=trim,
        digitize=digitize,
        all_consensus_cols=all_consensus_cols
    )


# add a very limited CLI so that this module can be invoked in a shell:
#     $ python -m pyhmmer.hmmsearch <hmmfile> <seqdb>
if __name__ == "__main__":

    import argparse
    import sys

    def _hmmsearch(args: argparse.Namespace) -> int:
        with SequenceFile(args.seqdb) as seqfile:
            alphabet = seqfile.guess_alphabet()
            if alphabet is None:
                print("could not guess alphabet of input, exiting", file=sys.stderr)
                return 1
            seqfile.set_digital(alphabet)
            sequences: typing.List[DigitalSequence] = list(seqfile)  # type: ignore

        with HMMFile(args.hmmfile) as hmms:
            queries = hmms.optimized_profiles() if hmms.is_pressed() else hmms
            hits_list = hmmsearch(queries, sequences, cpus=args.jobs)
            for hits in hits_list:
                for hit in hits:
                    if hit.is_reported():
                        print(
                            hit.name.decode(),
                            "-",
                            hit.best_domain.alignment.hmm_accession.decode(),
                            hit.best_domain.alignment.hmm_name.decode(),
                            hit.evalue,
                            hit.score,
                            hit.bias,
                            sep="\t",
                        )

        return 0

    def _phmmer(args: argparse.Namespace) -> int:
        with SequenceFile(args.seqdb) as seqfile:
            alphabet = seqfile.guess_alphabet() or Alphabet.amino()
            seqfile.set_digital(alphabet)
            sequences = list(seqfile)

        with SequenceFile(args.seqfile) as queries:
            queries.set_digital(alphabet)
            hits_list = phmmer(queries, sequences, cpus=args.jobs)  # type: ignore

            for hits in hits_list:
                for hit in hits:
                    if hit.is_reported():
                        print(
                            hit.name.decode(),
                            "-",
                            hit.best_domain.alignment.hmm_accession.decode(),
                            hit.best_domain.alignment.hmm_name.decode(),
                            hit.evalue,
                            hit.score,
                            hit.bias,
                            sep="\t",
                        )

        return 0

    def _nhmmer(args: argparse.Namespace) -> int:
        with SequenceFile(args.seqdb) as seqfile:
            alphabet = seqfile.guess_alphabet() or Alphabet.dna()
            seqfile.set_digital(alphabet)
            sequences = list(seqfile)

        with SequenceFile(args.seqfile) as queryfile:
            queryfile.set_digital(alphabet)
            queries = list(queryfile)
            hits_list = nhmmer(queries, sequences, cpus=args.jobs)  # type: ignore
            for hits in hits_list:
                for hit in hits:
                    if hit.is_reported():
                        print(
                            hit.name.decode(),
                            "-",
                            hit.best_domain.alignment.hmm_accession.decode(),
                            hit.best_domain.alignment.hmm_name.decode(),
                            hit.evalue,
                            hit.score,
                            hit.bias,
                            sep="\t",
                        )

        return 0

    def _hmmpress(args: argparse.Namespace) -> int:
        for ext in ["h3m", "h3i", "h3f", "h3p"]:
            path = "{}.{}".format(args.hmmfile, ext)
            if os.path.exists(path):
                if args.force:
                    os.remove(path)
                else:
                    print(f"file {path} already exists")
                    return 1

        with HMMFile(args.hmmfile) as hmms:
            hmmpress(hmms, args.hmmfile)

        return 0

    def _hmmalign(args: argparse.Namespace) -> int:
        with SequenceFile(args.seqfile, format=args.informat) as seqfile:
            alphabet = seqfile.guess_alphabet()
            if alphabet is None:
                print("could not guess alphabet of input, exiting", file=sys.stderr)
                return 1
            seqfile.set_digital(alphabet)
            sequences: typing.List[DigitalSequence] = list(seqfile)  # type: ignore

        with HMMFile(args.hmmfile) as hmms:
            hmm = next(hmms)
            if next(hmms, None) is not None:
                print("HMM file contains more than one HMM, exiting", file=sys.stderr)
                return 1

        msa = hmmalign(hmm, sequences, trim=args.trim)
        if args.output == "-":
            with io.BytesIO() as out:
                msa.write(out, args.outformat)
                print(out.getvalue().decode("ascii"), end="")
        else:
            with open(args.output, "wb") as out:  # type: ignore
                msa.write(out, args.outformat)

        return 0

    parser = argparse.ArgumentParser()
    parser.add_argument("-j", "--jobs", required=False, default=0, type=int)
    subparsers = parser.add_subparsers(
        dest="cmd", help="HMMER command to run", required=True
    )

    parser_hmmsearch = subparsers.add_parser("hmmsearch")
    parser_hmmsearch.set_defaults(call=_hmmsearch)
    parser_hmmsearch.add_argument("hmmfile")
    parser_hmmsearch.add_argument("seqdb")

    parser_phmmer = subparsers.add_parser("phmmer")
    parser_phmmer.set_defaults(call=_phmmer)
    parser_phmmer.add_argument("seqfile")
    parser_phmmer.add_argument("seqdb")

    parser_nhmmer = subparsers.add_parser("nhmmer")
    parser_nhmmer.set_defaults(call=_nhmmer)
    parser_nhmmer.add_argument("seqfile")
    parser_nhmmer.add_argument("seqdb")

    parser_hmmpress = subparsers.add_parser("hmmpress")
    parser_hmmpress.set_defaults(call=_hmmpress)
    parser_hmmpress.add_argument("hmmfile")
    parser_hmmpress.add_argument("-f", "--force", action="store_true")

    parser_hmmalign = subparsers.add_parser("hmmalign")
    parser_hmmalign.set_defaults(call=_hmmalign)
    parser_hmmalign.add_argument(
        "hmmfile",
        metavar="<hmmfile>"
    )
    parser_hmmalign.add_argument(
        "seqfile",
        metavar="<seqfile>",
    )
    parser_hmmalign.add_argument(
        "-o",
        "--output",
        action="store",
        default="-",
        metavar="<f>",
        help="output alignment to file <f>, not stdout"
    )
    parser_hmmalign.add_argument(
        "--trim",
        action="store_true",
        help="trim terminal tails of nonaligned residues from alignment"
    )
    parser_hmmalign.add_argument(
        "--informat",
        action="store",
        metavar="<s>",
        help="assert <seqfile> is in format <s> (no autodetection)",
        choices=SequenceFile._FORMATS.keys(),
    )
    parser_hmmalign.add_argument(
        "--outformat",
        action="store",
        metavar="<s>",
        help="output alignment in format <s>",
        default="stockholm",
        choices=MSAFile._FORMATS.keys(),
    )

    args = parser.parse_args()
    sys.exit(args.call(args))



from libhmmer.p7_hmm cimport P7_HMM
from libhmmer.p7_profile cimport P7_PROFILE
from libhmmer.p7_bg cimport P7_BG


cdef extern from "hmmer.h" nogil:
    extern int p7_ProfileConfig(const P7_HMM *hmm, const P7_BG *bg, P7_PROFILE *gm, int L, int mode);
    extern int p7_ReconfigLength  (P7_PROFILE *gm, int L);
    extern int p7_ReconfigMultihit(P7_PROFILE *gm, int L);
    extern int p7_ReconfigUnihit  (P7_PROFILE *gm, int L);
